import os
from pathlib import Path

from .args import parse_args
from .aur import find_aur_packages, get_repo_url
from .aur_deps import get_aur_deps_list
from .core import check_runtime_deps, interactive_spawn
from .exceptions import PackagesNotFoundInRepoError
from .i18n import translate
from .pacman import PackageDB
from .pprint import print_stdout
from .print_department import print_not_found_packages
from .urllib import wrap_proxy_env


def cli_getpkgbuild() -> None:
    args = parse_args()
    pwd = Path(os.path.curdir).resolve()
    aur_pkg_names = args.positional

    aur_pkgs, not_found_aur_pkgs = find_aur_packages(aur_pkg_names)
    repo_pkgs = []
    not_found_repo_pkgs = []
    for pkg_name in not_found_aur_pkgs:
        try:
            repo_pkg = PackageDB.find_repo_package(pkg_name)
        except PackagesNotFoundInRepoError:
            not_found_repo_pkgs.append(pkg_name)
        else:
            repo_pkgs.append(repo_pkg)

    if repo_pkgs:
        check_runtime_deps(["pkgctl"])

    if not_found_repo_pkgs:
        print_not_found_packages(not_found_repo_pkgs)

    if args.deps:
        aur_pkgs = aur_pkgs + get_aur_deps_list(aur_pkgs)

    for aur_pkg in aur_pkgs:
        name = aur_pkg.name
        repo_path = pwd / name
        print_stdout()
        interactive_spawn(wrap_proxy_env([
            "git",
            "clone",
            get_repo_url(aur_pkg.packagebase),
            str(repo_path),
        ]))

    for repo_pkg in repo_pkgs:
        name = repo_pkg.name
        repo_path = pwd / name
        action = "clone"
        if repo_path.exists():
            action = "update"
        print_stdout()
        print_stdout(translate(f"Package '{name}' going to be cloned into '{repo_path}'..."))
        if action == "clone":
            interactive_spawn([
                "pkgctl",
                "repo",
                "clone",
                "--unprivileged",
                name,
            ])
        elif action == "update":
            interactive_spawn([
                "git",
                "-C", repo_path.as_posix(),
                "pull",
                "origin",
                "main",
            ])
        else:
            raise NotImplementedError



"""Licensed under GPLv3, see https://www.gnu.org/licenses/"""

import sys
from threading import Lock
from typing import TYPE_CHECKING

from .pprint import color_enabled, get_term_width

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Any, Final


class ProgressBar():

    print_ratio: float
    index = 0
    progress = 0

    LEFT_DECORATION: "Final" = "["
    RIGHT_DECORATION: "Final" = "]"
    EMPTY: "Final" = "-"
    FULL: "Final" = "#"

    def __init__(self, length: int, message: str = "") -> None:
        width = (
            get_term_width() - len(message) - len(self.LEFT_DECORATION) - len(self.RIGHT_DECORATION)
        )
        self.print_ratio = length / width
        sys.stderr.write(message)
        if color_enabled():
            sys.stderr.write(self.LEFT_DECORATION + self.EMPTY * width + self.RIGHT_DECORATION)
            sys.stderr.write(f"{(chr(27))}[1D" * (width + len(self.RIGHT_DECORATION)))

    def update(self) -> None:
        self.index += 1
        if self.index / self.print_ratio > self.progress:
            self.progress += 1
            if color_enabled():
                sys.stderr.write(self.FULL)

    def __enter__(self) -> "Callable[[], None]":
        return self.update

    def __exit__(self, *_exc_details: "Any") -> None:
        sys.stderr.write("\n")


class ThreadSafeProgressBar():

    _progressbar_storage: dict[str, ProgressBar] = {}
    _progressbar_lock = Lock()

    @classmethod
    def get(cls, progressbar_length: int, progressbar_id: str) -> ProgressBar:
        if progressbar_id not in cls._progressbar_storage:
            with cls._progressbar_lock:
                if progressbar_id not in cls._progressbar_storage:
                    sys.stderr.write("\n")
                    cls._progressbar_storage[progressbar_id] = ProgressBar(
                        length=progressbar_length,
                        message="Synchronizing AUR database... ",
                    )
        return cls._progressbar_storage[progressbar_id]



"""Licensed under GPLv3, see https://www.gnu.org/licenses/"""

import configparser
import os
import random
import sys
from pathlib import Path
from tempfile import gettempdir
from typing import TYPE_CHECKING

from .i18n import PIKAUR_NAME, translate

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Any, Final

    from typing_extensions import NotRequired, TypedDict

    class DeprecatedConfigValue(TypedDict):
        section: str
        option: str
        transform: NotRequired[Callable[[str, configparser.ConfigParser], str]]

    class ConfigValueType(TypedDict):
        data_type: str
        default: NotRequired[str]
        deprecated: NotRequired[DeprecatedConfigValue]
        migrated: NotRequired[bool]


DEFAULT_CONFIG_ENCODING: "Final" = "utf-8"
BOOL: "Final" = "bool"
INT: "Final" = "int"
STR: "Final" = "str"
RUNNING_AS_ROOT: "Final" = os.geteuid() == 0  # @TODO: could global var be avoided here?
VERSION: "Final" = "1.15.2-dev"

_USER_TEMP_ROOT: "Final" = Path(gettempdir())
_USER_CACHE_ROOT: "Final" = Path(os.environ.get(
    "XDG_CACHE_HOME",
    Path.home() / ".cache/",
))

CACHE_ROOT: "Final" = (
    Path("/var/cache/pikaur")
    if RUNNING_AS_ROOT else
    _USER_CACHE_ROOT / "pikaur/"
)

BUILD_CACHE_PATH: "Final" = CACHE_ROOT / "build"
PACKAGE_CACHE_PATH: "Final" = CACHE_ROOT / "pkg"

CONFIG_ROOT: "Final" = Path(os.environ.get(
    "XDG_CONFIG_HOME",
    Path.home() / ".config/",
))

DATA_ROOT: "Final" = (
    Path(os.environ.get(
        "XDG_DATA_HOME",
        Path.home() / ".local/share/",
    )) / "pikaur"
)

_OLD_AUR_REPOS_CACHE_PATH: "Final" = CACHE_ROOT / "aur_repos"
AUR_REPOS_CACHE_PATH: "Final" = (
    (CACHE_ROOT / "aur_repos")
    if RUNNING_AS_ROOT else
    (DATA_ROOT / "aur_repos")
)

BUILD_DEPS_LOCK: "Final" = (
    (
        _USER_CACHE_ROOT if RUNNING_AS_ROOT else _USER_TEMP_ROOT)
    / "pikaur_build_deps.lock"
)
PROMPT_LOCK: "Final" = (
    (
        _USER_CACHE_ROOT if RUNNING_AS_ROOT else _USER_TEMP_ROOT
    ) / f"pikaur_prompt_{random.randint(0, 999999)}.lock"  # nosec: B311   # noqa: S311
)


def get_config_path() -> Path:
    config_flag = "--pikaur-config"
    if config_flag in sys.argv:
        return Path(sys.argv[
            sys.argv.index(config_flag) + 1
        ])
    return CONFIG_ROOT / "pikaur.conf"


class UpgradeSortingValues:
    VERSIONDIFF: "Final" = "versiondiff"
    PKGNAME: "Final" = "pkgname"
    REPO: "Final" = "repo"


class AurSearchSortingValues:
    HOTTEST: "Final" = "hottest"
    PKGNAME: "Final" = "pkgname"
    POPULARITY: "Final" = "popularity"
    NUMVOTES: "Final" = "numvotes"
    LASTMODIFIED: "Final" = "lastmodified"


class DiffPagerValues:
    AUTO: "Final" = "auto"
    ALWAYS: "Final" = "always"
    NEVER: "Final" = "never"


CONFIG_YES_VALUES: "Final" = ("yes", "y", "true", "1")


ConfigSchemaT = dict[str, dict[str, "ConfigValueType"]]
CONFIG_SCHEMA: ConfigSchemaT = {
    "sync": {
        "AlwaysShowPkgOrigin": {
            "data_type": BOOL,
            "default": "no",
        },
        "DevelPkgsExpiration": {
            "data_type": INT,
            "default": "-1",
        },
        "UpgradeSorting": {
            "data_type": STR,
            "default": UpgradeSortingValues.VERSIONDIFF,
        },
        "ShowDownloadSize": {
            "data_type": BOOL,
            "default": "no",
        },
        "IgnoreOutofdateAURUpgrades": {
            "data_type": BOOL,
            "default": "no",
        },
    },
    "build": {
        "KeepBuildDir": {
            "data_type": BOOL,
            "default": "no",
        },
        "KeepDevBuildDir": {
            "data_type": BOOL,
            "default": "yes",
        },
        "KeepBuildDeps": {
            "data_type": BOOL,
            "default": "no",
        },
        "GpgDir": {
            "data_type": STR,
            "default": ("/etc/pacman.d/gnupg/" if RUNNING_AS_ROOT else ""),
        },
        "SkipFailedBuild": {
            "data_type": BOOL,
            "default": "no",
        },
        "DynamicUsers": {
            "data_type": STR,
            "default": "root",
        },
        "AlwaysUseDynamicUsers": {
            "data_type": BOOL,
            "default": "no",
            "deprecated": {
                "section": "build",
                "option": "DynamicUsers",
                "transform": lambda old_value, _config: "always" if old_value == "yes" else "root",
            },
        },
        "NoEdit": {
            "data_type": BOOL,
            "deprecated": {
                "section": "review",
                "option": "NoEdit",
            },
        },
        "DontEditByDefault": {
            "data_type": BOOL,
            "deprecated": {
                "section": "review",
                "option": "DontEditByDefault",
            },
        },
        "NoDiff": {
            "data_type": BOOL,
            "deprecated": {
                "section": "review",
                "option": "NoDiff",
            },
        },
        "GitDiffArgs": {
            "data_type": STR,
            "deprecated": {
                "section": "review",
                "option": "GitDiffArgs",
            },
        },
        "IgnoreArch": {
            "data_type": BOOL,
            "default": "no",
        },
    },
    "review": {
        "NoEdit": {
            "data_type": BOOL,
            "default": "no",
        },
        "DontEditByDefault": {
            "data_type": BOOL,
            "default": "no",
        },
        "NoDiff": {
            "data_type": BOOL,
            "default": "no",
        },
        "GitDiffArgs": {
            "data_type": STR,
            "default": "--ignore-space-change,--ignore-all-space",
        },
        "DiffPager": {
            "data_type": STR,
            "default": DiffPagerValues.AUTO,
        },
        "HideDiffFiles": {
            "data_type": STR,
            "default": ".SRCINFO",
        },
    },
    "colors": {
        "Version": {
            "data_type": INT,
            "default": "10",
        },
        "VersionDiffOld": {
            "data_type": INT,
            "default": "11",
        },
        "VersionDiffNew": {
            "data_type": INT,
            "default": "9",
        },
    },
    "ui": {
        "RequireEnterConfirm": {
            "data_type": BOOL,
            "default": "yes",
        },
        "DiffPager": {
            "data_type": STR,
            "deprecated": {
                "section": "review",
                "option": "DiffPager",
            },
        },
        "PrintCommands": {
            "data_type": BOOL,
            "default": "no",
        },
        "AurSearchSorting": {
            "data_type": STR,
            "default": AurSearchSortingValues.HOTTEST,
        },
        "DisplayLastUpdated": {
            "data_type": BOOL,
            "default": "no",
        },
        "GroupByRepository": {
            "data_type": BOOL,
            "default": "yes",
        },
        "ReverseSearchSorting": {
            "data_type": BOOL,
            "default": "no",
        },
        "WarnAboutPackageUpdates": {
            "data_type": STR,
            "default": "",
        },
    },
    "misc": {
        "SudoLoopInterval": {
            "data_type": INT,
            "default": "59",
        },
        "PacmanPath": {
            "data_type": STR,
            "default": "pacman",
        },
        "PrivilegeEscalationTool": {
            "data_type": STR,
            "default": "sudo",
        },
        "AurHost": {
            "data_type": STR,
            "deprecated": {
                "section": "network",
                "option": "AurUrl",
                "transform": lambda old_value, _config: f"https://{old_value}",
            },
        },
        "NewsUrl": {
            "data_type": STR,
            "deprecated": {
                "section": "network",
                "option": "NewsUrl",
            },
        },
    },
    "network": {
        "AurUrl": {
            "data_type": STR,
            "default": "https://aur.archlinux.org",
        },
        "NewsUrl": {
            "data_type": STR,
            "default": "https://www.archlinux.org/feeds/news/",
        },
        "Socks5Proxy": {
            "data_type": STR,
            "default": "",
        },
        "AurHttpProxy": {
            "data_type": STR,
            "default": "",
        },
        "AurHttpsProxy": {
            "data_type": STR,
            "default": "",
        },
    },
}


def get_key_type(section_name: str, key_name: str) -> str | None:
    config_value: "ConfigValueType" | None = CONFIG_SCHEMA.get(section_name, {}).get(key_name, None)
    if not config_value:
        return None
    return config_value.get("data_type")


def write_config(config: configparser.ConfigParser | None = None) -> None:
    if not config:
        config = configparser.ConfigParser()
    need_write = False
    for section_name, section in CONFIG_SCHEMA.items():
        if section_name not in config:
            config[section_name] = {}
        for option_name, option_schema in section.items():
            if option_schema.get("migrated"):
                need_write = True
                continue
            if option_schema.get("deprecated"):
                continue
            if option_name not in config[section_name]:
                config[section_name][option_name] = option_schema["default"]
                need_write = True
    if need_write:
        if not CONFIG_ROOT.exists():
            CONFIG_ROOT.mkdir(parents=True)
        with get_config_path().open("w", encoding=DEFAULT_CONFIG_ENCODING) as configfile:
            config.write(configfile)


def str_to_bool(value: str) -> bool:
    return value.lower() in CONFIG_YES_VALUES


class PikaurConfigItem:

    def __init__(self, section: configparser.SectionProxy, key: str) -> None:
        self.section = section
        self.key = key
        self.value = self.section.get(key)
        self._type_error_template = translate(
            "{key} is not '{typeof}'",
        )

    def get_bool(self) -> bool:
        if get_key_type(self.section.name, self.key) != BOOL:
            not_bool_error = self._type_error_template.format(key=self.key, typeof=BOOL)
            raise TypeError(not_bool_error)
        return str_to_bool(self.value)

    def get_int(self) -> int:
        if get_key_type(self.section.name, self.key) != INT:
            not_int_error = self._type_error_template.format(key=self.key, typeof=INT)
            raise TypeError(not_int_error)
        return int(self.value)

    def get_str(self) -> str:
        # note: it"s basically needed for mypy
        if get_key_type(self.section.name, self.key) != STR:
            not_str_error = self._type_error_template.format(key=self.key, typeof=STR)
            raise TypeError(not_str_error)
        return str(self.value)

    def __str__(self) -> str:
        return self.get_str()

    def __eq__(self, item: "Any") -> bool:
        result: bool = self.get_str() == item
        return result


class PikaurConfigSection():

    section: configparser.SectionProxy

    def __init__(self, section: configparser.SectionProxy) -> None:
        self.section = section

    def __getattr__(self, attr: str) -> PikaurConfigItem:
        return PikaurConfigItem(self.section, attr)

    def __repr__(self) -> str:
        return str(self.section)


class PikaurConfig():

    _config: configparser.ConfigParser

    @classmethod
    def get_config(cls) -> configparser.ConfigParser:
        if not getattr(cls, "_config", None):
            config_path = get_config_path()
            cls._config = configparser.ConfigParser()
            if not config_path.exists():
                write_config()
            cls._config.read(config_path, encoding=DEFAULT_CONFIG_ENCODING)
            cls.migrate_config()
            write_config(config=cls._config)
            cls.validate_config()
        return cls._config

    @classmethod
    def migrate_config(cls) -> None:
        for section_name, section in CONFIG_SCHEMA.items():
            for option_name, option_schema in section.items():
                if not option_schema.get("deprecated"):
                    continue

                new_section_name: str = option_schema["deprecated"]["section"]
                new_option_name: str = option_schema["deprecated"]["option"]
                transform: """Callable[
                    [str, configparser.ConfigParser], str,
                ] | None""" = option_schema["deprecated"].get("transform")

                old_value_was_migrated = False
                value_to_migrate = None
                if (
                        new_section_name not in cls._config
                ) or (
                    cls._config[new_section_name].get(new_option_name) is None
                ):
                    value_to_migrate = cls._config[section_name].get(option_name)
                    if value_to_migrate is not None:
                        if transform:
                            new_value = transform(value_to_migrate, cls._config)
                        else:
                            new_value = value_to_migrate
                        if new_section_name not in cls._config:
                            cls._config[new_section_name] = {}
                        cls._config[new_section_name][new_option_name] = new_value
                        CONFIG_SCHEMA[new_section_name][new_option_name]["migrated"] = True
                        old_value_was_migrated = True

                old_value_was_removed = False
                if option_name in cls._config[section_name]:
                    del cls._config[section_name][option_name]
                    CONFIG_SCHEMA[section_name][option_name]["migrated"] = True
                    old_value_was_removed = True

                if old_value_was_migrated or old_value_was_removed:
                    print(" ".join([  # noqa: T201
                        "::",
                        translate("warning:"),
                        translate('Migrating [{}]{}="{}" config option to [{}]{}="{}"...').format(
                            section_name, option_name,
                            value_to_migrate or "",
                            new_section_name, new_option_name,
                            cls._config[new_section_name][new_option_name],
                        ),
                        "\n",
                    ]))

    @classmethod
    def validate_config(cls) -> None:
        pacman_path = cls._config["misc"]["PacmanPath"]
        if pacman_path in (PIKAUR_NAME, sys.argv[0]) or pacman_path.endswith(f"/{PIKAUR_NAME}"):
            print("BAM! I am a shell bomb.")  # noqa: T201
            sys.exit(1)

    def __getattr__(self, attr: str) -> PikaurConfigSection:
        return PikaurConfigSection(self.get_config()[attr])



"""Licensed under GPLv3, see https://www.gnu.org/licenses/"""

from logging import Logger
from typing import TYPE_CHECKING

from .args import parse_args
from .i18n import translate
from .pprint import Colors, ColorsHighlight, color_line, print_stderr

if TYPE_CHECKING:
    from typing import Any


class DebugColorCounter:

    # cyan is purposely skipped as it's used in print_debug itself,
    # highlight-red is purposely skipped as it's used in print_error,
    # highlight-yellow is purposely skipped as it's used in print_warning:
    colors = [
        Colors.red,
        Colors.green,
        Colors.yellow,
        Colors.blue,
        Colors.purple,
        Colors.white,
        ColorsHighlight.green,
        ColorsHighlight.blue,
        ColorsHighlight.purple,
        ColorsHighlight.cyan,
        ColorsHighlight.white,
    ]
    _current_color_idx = 0

    @classmethod
    def get_next(cls) -> int:
        color = cls.colors[cls._current_color_idx]
        cls._current_color_idx += 1
        if cls._current_color_idx >= len(cls.colors):
            cls._current_color_idx = 0
        return color


def print_debug(message: "Any", *, lock: bool = True) -> None:
    args = parse_args()
    if not args.pikaur_debug:
        return
    prefix = translate("debug:")
    if args.debug:
        # to avoid mixing together with pacman's debug messages:
        prefix = translate("pikaur debug:")
    print_stderr(" ".join([
        color_line(":: " + prefix, Colors.cyan),
        str(message),
    ]), lock=lock)


class PikaurLogger(Logger):  # we inherit `Logger` class only for pylint warnings to catch up on it
    def __init__(  # pylint: disable=super-init-not-called
            self, module_name: str, color: int, lock: bool | None = None,
    ) -> None:
        self.module_name = module_name
        self.color = color
        self.parent_lock = lock

    def debug(
        self, msg: "Any", *args: "Any", lock: bool | None = None, **kwargs: "Any",
    ) -> None:
        lock = lock if (lock is not None) else self.parent_lock
        str_message = msg.format(*args, **kwargs) if isinstance(msg, str) else str(msg)
        msg = f"{color_line(self.module_name, self.color)}: {str_message}"
        if lock is not None:
            print_debug(msg, lock=lock)
        else:
            print_debug(msg)


def create_logger(
        module_name: str, *, lock: bool | None = None,
) -> PikaurLogger:
    return PikaurLogger(module_name=module_name, color=DebugColorCounter.get_next(), lock=lock)



"""Licensed under GPLv3, see https://www.gnu.org/licenses/"""

import sys
from multiprocessing.pool import ThreadPool
from typing import TYPE_CHECKING

from .args import parse_args
from .aur import (
    AURPackageInfo,
    AurRPCErrors,
    aur_rpc_search_name_desc,
    get_all_aur_names,
    get_all_aur_packages,
)
from .exceptions import AURError, SysExit
from .i18n import translate
from .pacman import PackageDB, get_pkg_id, refresh_pkg_db_if_needed
from .pprint import print_error, print_stderr
from .print_department import print_package_search_results

if TYPE_CHECKING:
    from collections.abc import Iterable
    from typing import TypeVar

    import pyalpm

    from .print_department import AnyPackage

    SamePackageTypeT = TypeVar("SamePackageTypeT", AURPackageInfo, pyalpm.Package)


def package_search_thread_repo(query: str) -> "list[pyalpm.Package]":
    args = parse_args()
    result = (
        PackageDB.search_repo(query, names_only=args.namesonly)
        if query else
        PackageDB.get_repo_list(quiet=True)
    )
    if not args.quiet:
        sys.stderr.write("#")
    return result


def filter_aur_results(
        results: dict[str, list[AURPackageInfo]],
        query: str,
        *,
        names_only: bool = False,
) -> dict[str, list[AURPackageInfo]]:
    filtered_results: dict[str, list[AURPackageInfo]] = {}
    for _q, pkgs in results.items():
        for pkg in pkgs:
            if (
                    query in pkg.name
            ) or (
                not names_only and
                (query in (pkg.desc or ""))
            ):
                filtered_results.setdefault(_q, []).append(pkg)
    return filtered_results


def package_search_thread_aur(  # pylint: disable=too-many-branches
        queries: list[str],
) -> list[AURPackageInfo]:
    args = parse_args()
    result = {}
    if queries:
        use_as_filters: list[str] = []
        with ThreadPool() as pool:
            requests = {}
            for query in queries:
                requests[query] = pool.apply_async(aur_rpc_search_name_desc, (query, ))
            pool.close()
            for query, request in requests.items():
                try:
                    result[query] = request.get()
                except AURError as exc:
                    if exc.error == AurRPCErrors.TOO_MANY_RESULTS:
                        print_error(
                            translate("AUR: Too many package results for '{query}'").format(
                                query=query,
                            ),
                        )
                        use_as_filters.append(query)
                    elif exc.error == AurRPCErrors.QUERY_TOO_SMALL:
                        print_error(
                            translate("AUR: Query arg too small '{query}'").format(
                                query=query,
                            ),
                        )
                        use_as_filters.append(query)
                    else:
                        raise
            pool.join()
        for query in use_as_filters:
            result = filter_aur_results(result, query, names_only=args.namesonly)
        if args.namesonly:
            for subindex, subresult in result.items():
                result[subindex] = [
                    pkg for pkg in subresult
                    if subindex in pkg.name
                ]
    else:
        if args.quiet:  # noqa: PLR5501
            result = {"all": [
                AURPackageInfo(
                    name=name,
                    packagebase=name,
                    version="0",
                ) for name in get_all_aur_names()
            ]}
        else:
            result = {"all": get_all_aur_packages()}
    if not args.quiet:
        sys.stderr.write("#")
    return list(join_search_results(list(result.values())))


def package_search_thread_local() -> dict[str, str]:
    result = {
        pkg_name: pkg.version
        for pkg_name, pkg in PackageDB.get_local_dict(quiet=True).items()
    }
    if not parse_args().quiet:
        sys.stderr.write("#")
    return result


def join_search_results(
        all_search_results: "list[list[SamePackageTypeT]]",
) -> "Iterable[SamePackageTypeT]":
    if not all_search_results:
        return []
    pkgnames_set: set[str] = set()
    for search_results in all_search_results:
        new_pkgnames_set = {get_pkg_id(result) for result in search_results}
        pkgnames_set = (
            pkgnames_set.intersection(new_pkgnames_set)
            if pkgnames_set else
            new_pkgnames_set
        )
    return {
        get_pkg_id(result): result
        for result in all_search_results[0]
        if get_pkg_id(result) in pkgnames_set
    }.values()


def search_packages(  # pylint: disable=too-many-locals
        *, enumerated: bool = False,
) -> "list[AnyPackage]":
    refresh_pkg_db_if_needed()

    args = parse_args()
    search_query = args.positional or []
    repo_only = args.repo
    aur_only = args.aur

    if not args.quiet:
        progressbar_length = max(len(search_query), 1) + (not repo_only) + (not aur_only)
        print_stderr(translate("Searching... [{bar}]").format(bar="-" * progressbar_length), end="")
        print_stderr("\x1b[1D" * (progressbar_length + 1), end="")

    with ThreadPool() as pool:
        request_local = pool.apply_async(package_search_thread_local, ())
        requests_repo = [
            pool.apply_async(package_search_thread_repo, (search_word, ))
            for search_word in (search_query or [""])
        ] if not aur_only else []
        request_aur = pool.apply_async(
            package_search_thread_aur, (search_query,),
        ) if not repo_only else None
        pool.close()

        result_local = request_local.get()
        result_repo: "list[list[pyalpm.Package]]" = []
        for request_repo in requests_repo:
            pkgs_found = request_repo.get()
            if pkgs_found:
                result_repo.append(pkgs_found)
        result_aur = None
        if request_aur:
            try:
                result_aur = request_aur.get()
            except AURError as exc:
                print_stderr(f"{translate('AUR returned error:')} {exc}")
                raise SysExit(121) from exc
        pool.join()

    if not args.quiet:
        sys.stderr.write("\n")

    joined_repo_results: "Iterable[pyalpm.Package]" = []
    if result_repo:
        joined_repo_results = join_search_results(result_repo)
    joined_aur_results: "Iterable[AURPackageInfo]" = result_aur or []

    return print_package_search_results(
        repo_packages=joined_repo_results,
        aur_packages=joined_aur_results,
        local_pkgs_versions=result_local,
        enumerated=enumerated,
    )


def cli_search_packages() -> None:
    search_packages()



"""Licensed under GPLv3, see https://www.gnu.org/licenses/"""

from datetime import datetime
from typing import TYPE_CHECKING

from .args import parse_args
from .aur import find_aur_packages
from .config import PikaurConfig
from .core import DEFAULT_TIMEZONE, AURInstallInfo, RepoInstallInfo
from .exceptions import PackagesNotFoundInRepoError
from .i18n import translate, translate_many
from .pacman import (
    PackageDB,
    PacmanConfig,
    find_packages_not_from_repo,
    find_upgradeable_packages,
    get_ignored_pkgnames_from_patterns,
)
from .pprint import print_stderr, print_stdout
from .print_department import (
    pretty_format_upgradeable,
    print_ignored_package,
    print_ignoring_outofdate_upgrade,
)
from .version import VERSION_DEVEL, compare_versions

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Final

    import pyalpm

    from .aur import AURPackageInfo
    from .core import InstallInfo


DEVEL_PKGS_POSTFIXES: "Final" = (
    "-git",
    "-svn",
    "-bzr",
    "-hg",
    "-cvs",
    "-nightly",
)


def is_devel_pkg(pkg_name: str) -> bool:
    result = False
    for devel_pkg_postfix in DEVEL_PKGS_POSTFIXES:
        if pkg_name.endswith(devel_pkg_postfix):
            result = True
            break
    return result


def get_remote_package(
        new_pkg_name: str,
) -> "pyalpm.Package | AURPackageInfo | None":
    try:
        repo_pkg = PackageDB.find_repo_package(new_pkg_name)
    except PackagesNotFoundInRepoError:
        aur_packages, _not_found = find_aur_packages([new_pkg_name])
        if aur_packages:
            return aur_packages[0]
        return None
    return repo_pkg


def get_remote_package_version(new_pkg_name: str) -> str | None:
    pkg = get_remote_package(new_pkg_name)
    if pkg:
        return pkg.version
    return None


def find_repo_upgradeable() -> list[RepoInstallInfo]:
    """
    Unlike `pikaur.install_info_fetcher.InstallInfoFetcher.get_upgradeable_repo_pkgs_info`
    it find all upgradeable repo packages, even ignored ones or conflicting
    (like `pacman -Qu`).
    """
    all_local_pkgs = PackageDB.get_local_dict()
    repo_packages_updates = []
    for repo_pkg in find_upgradeable_packages():
        local_pkg = all_local_pkgs[repo_pkg.name]
        repo_packages_updates.append(
            RepoInstallInfo(
                name=local_pkg.name,
                new_version=repo_pkg.version,
                current_version=local_pkg.version,
                description=repo_pkg.desc,
                repository=repo_pkg.db.name,
                package=repo_pkg,
            ),
        )
    return repo_packages_updates


def find_aur_devel_updates(
        aur_pkgs_info: "list[AURPackageInfo]",
        package_ttl_days: int,
) -> list[AURInstallInfo]:
    local_packages = PackageDB.get_local_dict()
    now = datetime.now(tz=DEFAULT_TIMEZONE)
    aur_updates = []
    for aur_pkg in sorted(aur_pkgs_info, key=lambda x: x.name):
        pkg_name = aur_pkg.name
        if not is_devel_pkg(pkg_name):
            continue
        local_pkg = local_packages[pkg_name]
        pkg_install_datetime = datetime.fromtimestamp(
            local_pkg.installdate, tz=DEFAULT_TIMEZONE,
        )
        pkg_age_days = (now - pkg_install_datetime).days
        if pkg_age_days >= package_ttl_days:
            aur_updates.append(AURInstallInfo(
                name=pkg_name,
                current_version=local_pkg.version,
                new_version=VERSION_DEVEL,
                description=aur_pkg.desc,
                devel_pkg_age_days=pkg_age_days,
                maintainer=aur_pkg.maintainer,
                package=aur_pkg,
            ))
    return aur_updates


def find_aur_updates() -> tuple[list[AURInstallInfo], list[str]]:
    args = parse_args()
    package_names = find_packages_not_from_repo()
    print_stderr(translate_many(
        "Reading AUR package info...",
        "Reading AUR packages info...",
        len(package_names),
    ))
    aur_pkgs_info, not_found_aur_pkgs = find_aur_packages(package_names)
    local_packages = PackageDB.get_local_dict()
    aur_updates = []
    aur_pkgs_up_to_date = []
    for aur_pkg in aur_pkgs_info:
        pkg_name = aur_pkg.name
        aur_version = aur_pkg.version
        current_version = local_packages[pkg_name].version
        compare_aur_pkg = compare_versions(current_version, aur_version)
        if compare_aur_pkg < 0:
            pkg_install_info = AURInstallInfo(
                name=pkg_name,
                new_version=aur_version,
                current_version=current_version,
                description=aur_pkg.desc,
                maintainer=aur_pkg.maintainer,
                package=aur_pkg,
            )
            if args.ignore_outofdate and aur_pkg.outofdate:
                print_ignoring_outofdate_upgrade(pkg_install_info)
                continue
            aur_updates.append(pkg_install_info)
        else:
            aur_pkgs_up_to_date.append(aur_pkg)
    if aur_pkgs_up_to_date:
        sync_config = PikaurConfig().sync
        devel_packages_expiration = sync_config.DevelPkgsExpiration.get_int()
        if args.devel:
            devel_packages_expiration = 0
        if devel_packages_expiration > -1:
            aur_updates += find_aur_devel_updates(
                aur_pkgs_up_to_date,
                package_ttl_days=devel_packages_expiration,
            )
    return aur_updates, not_found_aur_pkgs


def print_upgradeable(
        aur_install_infos: "Sequence[InstallInfo] | None" = None,
        *,
        ignored_only: bool = False,
) -> None:
    args = parse_args()
    updates: "list[InstallInfo]" = []
    if aur_install_infos is not None:
        updates += aur_install_infos
    elif not args.repo:
        aur_updates, _not_found_aur_pkgs = find_aur_updates()
        updates += aur_updates
    if not args.aur:
        updates += find_repo_upgradeable()
    if not updates:
        return
    pkg_names = [pkg.name for pkg in updates]
    manually_ignored_pkg_names = get_ignored_pkgnames_from_patterns(
        pkg_names,
        args.ignore,
    )
    config_ignored_pkg_names = get_ignored_pkgnames_from_patterns(
        pkg_names,
        PacmanConfig().options.get("IgnorePkg", []),
    )
    ignored_pkg_names = manually_ignored_pkg_names + config_ignored_pkg_names
    for pkg in updates[:]:
        if pkg.name in ignored_pkg_names:
            updates.remove(pkg)
            ignored_from = None
            if pkg.name in config_ignored_pkg_names:
                ignored_from = translate("(ignored in Pacman config)")
            print_ignored_package(install_info=pkg, ignored_from=ignored_from)
    if ignored_only:
        return
    if args.quiet:
        print_stdout("\n".join([
            pkg_update.name for pkg_update in updates
        ]))
    else:
        print_stdout(pretty_format_upgradeable(
            updates,
            print_repo=PikaurConfig().sync.AlwaysShowPkgOrigin.get_bool(),
        ))



from typing import TYPE_CHECKING

from .args import parse_args, reconstruct_args
from .config import BUILD_CACHE_PATH, PACKAGE_CACHE_PATH, PikaurConfig
from .core import interactive_spawn, remove_dir, sudo
from .exceptions import SysExit
from .i18n import translate
from .logging import create_logger
from .pikspect import YesNo, format_pacman_question, pikspect
from .pprint import ColorsHighlight, bold_line, color_line, print_stdout
from .prompt import ask_to_continue

if TYPE_CHECKING:
    from collections.abc import Callable
    from subprocess import Popen  # nosec B404


_debug = create_logger("pkg_cache_cli").debug


def clean_aur_cache() -> None:
    args = parse_args()
    for directory, message, minimal_clean_level in (
            (BUILD_CACHE_PATH, translate("Build directory"), 1),
            (PACKAGE_CACHE_PATH, translate("Packages directory"), 2),
    ):
        print_stdout(f"\n{message}: {directory}")
        if minimal_clean_level > args.clean:
            continue
        if not directory.exists():
            print_stdout(translate("Directory is empty."))
        elif ask_to_continue(text="{} {}".format(  # pylint: disable=consider-using-f-string
                color_line("::", ColorsHighlight.blue),
                bold_line(translate("Do you want to remove all files?")),
        )):
            print_stdout(translate("removing all files from cache..."))
            remove_dir(directory)


def clean_repo_cache() -> None:
    args = parse_args()
    spawn_func: "Callable[[list[str]], Popen[bytes]]" = interactive_spawn
    if args.noconfirm:

        def noconfirm_cache_remove(pacman_args: list[str]) -> "Popen[bytes]":
            return pikspect(pacman_args, extra_questions={YesNo.ANSWER_Y: [
                format_pacman_question(
                    "Do you want to remove ALL files from cache?",
                    question=YesNo.QUESTION_YN_NO,
                ),
                format_pacman_question(
                    "Do you want to remove all other packages from cache?",
                ),
                format_pacman_question(
                    "Do you want to remove unused repositories?",
                ),
            ]})
        spawn_func = noconfirm_cache_remove
    raise SysExit(
        spawn_func(sudo([
            PikaurConfig().misc.PacmanPath.get_str(),
            *reconstruct_args(args, ignore_args=["noconfirm"]),
        ])).returncode,
    )


def cli_clean_packages_cache() -> None:
    args = parse_args()
    if not args.repo:
        clean_aur_cache()
    if not args.aur:
        clean_repo_cache()



"""Licensed under GPLv3, see https://www.gnu.org/licenses/"""

import os
import shutil
from glob import glob
from multiprocessing.pool import ThreadPool
from pathlib import Path
from typing import TYPE_CHECKING

from .args import parse_args
from .aur import find_aur_packages, get_repo_url
from .config import (
    AUR_REPOS_CACHE_PATH,
    BUILD_CACHE_PATH,
    BUILD_DEPS_LOCK,
    PACKAGE_CACHE_PATH,
    PikaurConfig,
)
from .core import (
    PIPE,
    DataType,
    dirname,
    interactive_spawn,
    isolate_root_cmd,
    joined_spawn,
    open_file,
    remove_dir,
    replace_file,
    running_as_root,
    spawn,
    sudo,
)
from .exceptions import BuildError, CloneError, DependencyError, DependencyNotBuiltYetError, SysExit
from .filelock import FileLock
from .i18n import translate, translate_many
from .logging import create_logger
from .makepkg_config import MakePkgCommand, MakepkgConfig, get_pkgdest
from .pacman import PackageDB, get_pacman_command, install_built_deps
from .pprint import (
    ColorsHighlight,
    bold_line,
    color_enabled,
    color_line,
    print_error,
    print_stderr,
    print_stdout,
)
from .prompt import (
    ask_to_continue,
    get_editor_or_exit,
    get_input,
    retry_interactive_command_or_exit,
)
from .srcinfo import SrcInfo
from .updates import is_devel_pkg
from .urllib import wrap_proxy_env
from .version import VersionMatcher, compare_versions

if TYPE_CHECKING:
    from typing import Final

    from .args import PikaurArgs
    from .core import InteractiveSpawn, SpawnArgs
    from .pacman import ProvidedDependency

logger = create_logger("build")

DEFAULT_PKGBUILD_BASENAME: "Final" = "PKGBUILD"
ARCH_ANY: "Final" = "any"
IGNORE_PATHS_WHEN_COPYING: "Final[tuple[str]]" = (".git", )


class PkgbuildChanged(Exception):  # noqa: N818
    pass


def _shell(cmds: list[str]) -> "InteractiveSpawn":
    return interactive_spawn(isolate_root_cmd(wrap_proxy_env(cmds)))


def _mkdir(to_path: Path) -> None:
    mkdir_result = spawn(isolate_root_cmd(["mkdir", "-p", str(to_path)]))
    if mkdir_result.returncode != 0:
        print_stdout(mkdir_result.stdout_text)
        print_stderr(mkdir_result.stderr_text)
        raise RuntimeError(translate(f"Can't create destination directory '{to_path}'."))


def copy_aur_repo(from_path: Path, to_path: Path) -> None:
    from_path = from_path.resolve()
    to_path = to_path.resolve()
    if not to_path.exists():
        _mkdir(to_path)

    from_paths = []
    for src_path_str in glob(f"{from_path}/*") + glob(f"{from_path}/.*"):
        src_path = Path(src_path_str)
        if src_path.name not in IGNORE_PATHS_WHEN_COPYING:
            from_paths.append(src_path)
    to_path = to_path.parent / f"{to_path.name}/"

    cmd_args = isolate_root_cmd(["cp", "-r", *[str(path) for path in [*from_paths, to_path]]])

    result = spawn(cmd_args)
    if result.returncode != 0:
        if to_path.exists():
            remove_dir(to_path)
            _mkdir(to_path)
        result = interactive_spawn(cmd_args)
        if result.returncode != 0:
            raise RuntimeError(translate(f"Can't copy '{from_path}' to '{to_path}'."))


class PackageBuild(DataType):  # pylint: disable=too-many-public-methods
    # pylint: disable=too-many-instance-attributes
    clone = False
    pull = False

    package_base: str
    package_names: list[str]

    repo_path: Path
    pkgbuild_path: Path
    build_dir: Path
    build_gpgdir: str
    built_packages_paths: dict[str, Path]

    reviewed = False
    keep_build_dir = False
    skip_carch_check = False
    _source_repo_updated = False
    _build_files_copied = False

    failed: bool | None = None

    new_deps_to_install: list[str]
    new_make_deps_to_install: list[str]
    built_deps_to_install: dict[str, Path]

    args: "PikaurArgs"
    resolved_conflicts: list[list[str]] | None = None

    _local_pkgs_wo_build_deps: set[str]
    _local_pkgs_with_build_deps: set[str]
    _local_provided_pkgs_with_build_deps: dict[str, list["ProvidedDependency"]]

    def __repr__(self) -> str:
        return (
            f'<{self.__class__.__name__} "{self.package_base}" '
            f"{self.package_names}>"
        )

    def __init__(  # pylint: disable=super-init-not-called
            self,
            package_names: list[str] | None = None,
            pkgbuild_path: str | None = None,
    ) -> None:
        self.args = parse_args()

        if pkgbuild_path:
            self.repo_path = dirname(pkgbuild_path)
            self.pkgbuild_path = Path(pkgbuild_path)
            srcinfo = SrcInfo(pkgbuild_path=pkgbuild_path)
            srcinfo.regenerate()  # @TODO: this is a workaround for building
            # multiple PKGBUILDs from the same directory, find some better way.
            pkgbase = srcinfo.get_value("pkgbase")
            if pkgbase and srcinfo.pkgnames:
                self.package_names = package_names or srcinfo.pkgnames
                self.package_base = pkgbase
            else:
                no_pkgname_error = translate("Can't get package name from PKGBUILD")
                raise BuildError(no_pkgname_error)
        elif package_names:
            self.package_names = package_names
            self.package_base = find_aur_packages([package_names[0]])[0][0].packagebase
            self.repo_path = AUR_REPOS_CACHE_PATH / self.package_base
            self.pkgbuild_path = self.repo_path / DEFAULT_PKGBUILD_BASENAME
        else:
            missing_property_error = translate(
                "Either `{prop1}` or `{prop2}` should be set",
            ).format(
                prop1="package_names",
                prop2="pkgbuild_path",
            )
            raise NotImplementedError(missing_property_error)

        self.build_dir = BUILD_CACHE_PATH / self.package_base
        logger.debug("Build dir: {}", self.build_dir)
        self.build_gpgdir = self.args.build_gpgdir
        self.built_packages_paths = {}
        self.keep_build_dir = self.args.keepbuild or (
            is_devel_pkg(self.package_base) and PikaurConfig().build.KeepDevBuildDir.get_bool()
        )
        self.skip_carch_check = PikaurConfig().build.IgnoreArch.get_bool()

        if self.repo_path.exists():
            if (self.repo_path / ".git").exists():
                self.pull = True
            else:
                self.clone = True
        else:
            self.repo_path.mkdir(parents=True)
            self.clone = True

        self.reviewed = self.current_hash == self.last_installed_hash

        self._local_pkgs_wo_build_deps = set()
        self._local_pkgs_with_build_deps = set()
        self._local_provided_pkgs_with_build_deps = {}

    def git_reset_changed(self) -> "InteractiveSpawn":
        return _shell([
            "git",
            "-C", str(self.repo_path),
            "checkout",
            "--",
            "*",
        ])

    def git_stash(self) -> "InteractiveSpawn":
        return _shell([
            "git",
            "-C", str(self.repo_path),
            "stash",
        ])

    def git_stash_pop(self) -> "InteractiveSpawn":
        return _shell([
            "git",
            "-C", str(self.repo_path),
            "stash",
            "pop",
        ])

    def update_aur_repo(self) -> "InteractiveSpawn | None":
        cmd_args: list[str] | None = None
        if self.pull and not self.args.skip_aur_pull:
            cmd_args = [
                "git",
                "-C", str(self.repo_path),
                "pull",
                "origin",
                "master",
            ]
        if self.clone:
            cmd_args = [
                "git",
                "clone",
                get_repo_url(self.package_base),
                str(self.repo_path),
            ]
        result: "InteractiveSpawn | None" = None
        if cmd_args:
            result = spawn(isolate_root_cmd(wrap_proxy_env(cmd_args)))
        self.reviewed = self.current_hash == self.last_installed_hash
        return result

    @property
    def last_installed_file_path(self) -> Path:
        return self.repo_path / "last_installed.txt"

    @property
    def last_installed_hash(self) -> str | None:
        """Commit hash of AUR repo of last version of the pkg installed by Pikaur."""
        if self.last_installed_file_path.exists():
            with open_file(self.last_installed_file_path) as last_installed_file:
                lines = last_installed_file.readlines()
                if lines:
                    return lines[0].strip()
        return None

    def update_last_installed_file(self) -> None:
        git_hash_path = self.repo_path / ".git/refs/heads/master"
        if git_hash_path.exists():
            shutil.copy2(
                git_hash_path,
                self.last_installed_file_path,
            )

    @property
    def current_hash(self) -> str | None:
        """Commit hash of AUR repo of the pkg."""
        git_hash_path = self.repo_path / ".git/refs/heads/master"
        if not git_hash_path.exists():
            return None
        with open_file(git_hash_path) as current_hash_file:
            return current_hash_file.readlines()[0].strip()

    def get_latest_dev_sources(self, *, check_dev_pkgs: bool = True) -> None:
        if not self.reviewed:
            return
        self.prepare_build_destination()
        if (
                self._source_repo_updated
        ) or (
            not (is_devel_pkg(self.package_base) and check_dev_pkgs)
        ):
            return
        print_stdout("{} {}...".format(  # pylint: disable=consider-using-f-string
            color_line("::", ColorsHighlight.white),
            translate_many(
                "Downloading the latest sources for a devel package {}",
                "Downloading the latest sources for devel packages {}",
                len(self.package_names),
            ).format(
                bold_line(", ".join(self.package_names)),
            ),
        ))
        pkgver_result = joined_spawn(
            isolate_root_cmd(
                [*MakePkgCommand.get(), "--nobuild", "--nocheck", "--nodeps"],
                cwd=self.build_dir,
            ),
            cwd=self.build_dir,
        )
        if pkgver_result.returncode != 0:
            error_text = translate("failed to retrieve latest dev sources:")
            print_stderr()
            print_stderr()
            print_error(error_text)
            print_stderr(pkgver_result.stdout_text)
            if (
                    not self.args.skip_failed_build
            ) and (
                not ask_to_continue(default_yes=False)
            ):
                raise BuildError(error_text)
        SrcInfo(self.build_dir).regenerate()
        self._source_repo_updated = True

    def get_version(self, package_name: str) -> str:
        return SrcInfo(self.build_dir, package_name).get_version()

    def _compare_versions(self, compare_to: int) -> bool:
        local_db = PackageDB.get_local_dict()
        self.get_latest_dev_sources(check_dev_pkgs=self.args.needed)
        return min(
            compare_versions(
                local_db[pkg_name].version,
                self.get_version(pkg_name),
            ) == compare_to
            if pkg_name in local_db else False
            for pkg_name in self.package_names
        )

    @property
    def version_already_installed(self) -> bool:
        return self._compare_versions(0)

    @property
    def version_is_upgradeable(self) -> bool:
        return self._compare_versions(-1)

    @property
    def all_deps_to_install(self) -> list[str]:
        return self.new_make_deps_to_install + self.new_deps_to_install

    def _filter_built_deps(
            self,
            all_package_builds: dict[str, "PackageBuild"],
    ) -> None:

        def _mark_dep_resolved(dep: str) -> None:
            if dep in self.new_make_deps_to_install:
                self.new_make_deps_to_install.remove(dep)
            if dep in self.new_deps_to_install:
                self.new_deps_to_install.remove(dep)

        all_provided_pkgnames: dict[str, str] = {}
        for pkg_build in all_package_builds.values():
            for pkg_name in pkg_build.package_names:
                srcinfo = SrcInfo(
                    pkgbuild_path=pkg_build.pkgbuild_path, package_name=pkg_name,
                )
                all_provided_pkgnames.update({
                    provided_name: pkg_name
                    for provided_name in
                    [pkg_name, *srcinfo.get_values("provides")]
                })

        self.built_deps_to_install = {}

        for dep in self.all_deps_to_install:
            dep_name = VersionMatcher(dep).pkg_name
            if dep_name not in all_provided_pkgnames:
                continue
            package_build = all_package_builds[all_provided_pkgnames[dep_name]]
            if package_build == self:
                _mark_dep_resolved(dep)
                continue
            for pkg_name in package_build.package_names:
                if package_build.failed:
                    self.failed = True
                    raise DependencyError
                if not package_build.built_packages_paths.get(pkg_name):
                    raise DependencyNotBuiltYetError
                self.built_deps_to_install[pkg_name] = \
                    package_build.built_packages_paths[pkg_name]
                _mark_dep_resolved(dep)

    def _get_pacman_command(self, ignore_args: list[str] | None = None) -> list[str]:
        return get_pacman_command(ignore_args=ignore_args) + (
            ["--noconfirm"] if self.args.noconfirm else []
        )

    def install_built_deps(
            self,
            all_package_builds: dict[str, "PackageBuild"],
    ) -> None:

        self.get_deps(all_package_builds)
        if not self.built_deps_to_install:
            return

        print_stderr("{} {}:".format(  # pylint: disable=consider-using-f-string
            color_line("::", ColorsHighlight.purple),
            translate("Installing already built dependencies for {}").format(
                bold_line(", ".join(self.package_names))),
        ))

        try:
            update_self_deps = False
            for pkg_name, pkg_build in all_package_builds.items():
                if pkg_name in self.built_deps_to_install:
                    pkg_build.install_built_deps(all_package_builds)
                    update_self_deps = True
            if update_self_deps:
                self.get_deps(all_package_builds)
            install_built_deps(
                deps_names_and_paths=self.built_deps_to_install,
                resolved_conflicts=self.resolved_conflicts,
            )
        except DependencyError as dep_exc:
            self.failed = True
            raise dep_exc from None
        finally:
            PackageDB.discard_local_cache()

    def _set_built_package_path(self) -> None:
        pkg_paths_spawn = spawn(
            isolate_root_cmd(
                [*MakePkgCommand.get(), "--packagelist"],
                cwd=self.build_dir,
            ),
            cwd=self.build_dir,
        )
        if pkg_paths_spawn.returncode != 0:
            return
        if not pkg_paths_spawn.stdout_text:
            return
        logger.debug("Package names: {}", pkg_paths_spawn)
        pkg_paths = [Path(line) for line in pkg_paths_spawn.stdout_text.splitlines()]
        if not pkg_paths:
            return
        pkg_dest = get_pkgdest()
        logger.debug("PKGDEST: {}", pkg_dest)
        pkg_paths.sort(key=lambda x: len(str(x)))
        for pkg_name in self.package_names:
            pkg_path = pkg_paths[0]
            if len(pkg_paths) > 1:
                arch = MakepkgConfig.get("CARCH")
                for each_path in pkg_paths:
                    each_filename = Path(each_path).name
                    if pkg_name in each_filename and (
                            (f"-{arch}." in each_filename) or (f"-{ARCH_ANY}." in each_filename)
                    ):
                        pkg_path = Path(each_filename)
                        break
            pkg_basename = Path(pkg_path).name
            logger.debug("Full path: {}, base path: {}", pkg_path, pkg_basename)
            if pkg_path == Path(pkg_basename):
                pkg_path = (Path(pkg_dest) if pkg_dest else self.build_dir) / pkg_path
                logger.debug("Resolving full path: {} from base path: {}", pkg_path, pkg_basename)
            new_package_path = (
                Path(pkg_dest) if pkg_dest else PACKAGE_CACHE_PATH
            ) / pkg_basename
            logger.debug("New package path: {}", new_package_path)
            if not pkg_dest or MakePkgCommand.pkgdest_skipped:
                pkg_sig_path = pkg_path.parent / (pkg_path.name + ".sig")
                new_package_sig_path = new_package_path.parent / (
                    new_package_path.name + ".sig"
                )
                if not PACKAGE_CACHE_PATH.exists():
                    PACKAGE_CACHE_PATH.mkdir(parents=True)
                replace_file(pkg_path, new_package_path)
                replace_file(pkg_sig_path, new_package_sig_path)
            pkg_path = new_package_path
            if pkg_path and pkg_path.exists():
                self.built_packages_paths[pkg_name] = pkg_path

    def check_if_already_built(self) -> bool:
        self.get_latest_dev_sources()
        self._set_built_package_path()
        if (
                not self.args.rebuild and
                len(self.built_packages_paths) == len(self.package_names)
        ):
            print_stderr("{} {}\n".format(  # pylint: disable=consider-using-f-string
                color_line("::", ColorsHighlight.green),
                translate_many(
                    "Package {pkg} is already built. Pass '--rebuild' flag to force the build.",
                    "Packages {pkg} are already built. Pass '--rebuild' flag to force the build.",
                    len(self.package_names),
                ).format(
                    pkg=bold_line(", ".join(self.package_names)),
                ),
            ))
            return True
        return False

    def prepare_build_destination(self, *, flush: bool = False) -> None:
        if flush:
            remove_dir(self.build_dir)
            self._build_files_copied = False
        if self._build_files_copied:
            return
        if self.build_dir.exists() and not self.keep_build_dir:
            remove_dir(self.build_dir)
        copy_aur_repo(self.repo_path, self.build_dir)

        pkgbuild_name = self.pkgbuild_path.name
        if pkgbuild_name != DEFAULT_PKGBUILD_BASENAME:
            default_pkgbuild_path = self.build_dir / DEFAULT_PKGBUILD_BASENAME
            custom_pkgbuild_path = self.build_dir / pkgbuild_name
            if default_pkgbuild_path.exists():
                default_pkgbuild_path.unlink()
            os.renames(custom_pkgbuild_path, default_pkgbuild_path)
        self._build_files_copied = True

    def get_deps(
            self,
            all_package_builds: dict[str, "PackageBuild"],
            *,
            filter_built: bool = True,
            exclude_pkg_names: list[str] | None = None,
    ) -> None:
        exclude_pkg_names = exclude_pkg_names or []
        self.new_deps_to_install = []
        new_make_deps_to_install: list[str] = []
        new_check_deps_to_install: list[str] = []
        for package_name in self.package_names:
            if package_name in exclude_pkg_names:
                continue
            src_info = SrcInfo(pkgbuild_path=self.pkgbuild_path, package_name=package_name)
            for new_deps_version_matchers, deps_destination in (
                    (
                        src_info.get_build_depends(), self.new_deps_to_install,
                    ), (
                        src_info.get_build_makedepends(), new_make_deps_to_install,
                    ), (
                        src_info.get_build_checkdepends(), new_check_deps_to_install,
                    ),
            ):
                # find deps satisfied explicitly:
                new_deps_to_install = [
                    dep_line
                    for dep_lines in [
                        new_deps_version_matchers[dep_name].line.split(",")
                        for dep_name in PackageDB.get_not_found_local_packages([
                            vm.line for vm in new_deps_version_matchers.values()
                        ])
                    ]
                    for dep_line in dep_lines
                ]
                deps_destination += new_deps_to_install  # noqa: PLW2901
        self.new_make_deps_to_install = list(set(
            new_make_deps_to_install + new_check_deps_to_install,
        ))
        if filter_built:
            self._filter_built_deps(all_package_builds)

    def _install_repo_deps(self) -> None:
        if not self.all_deps_to_install:
            return

        print_stderr("{} {}:".format(  # pylint: disable=consider-using-f-string
            color_line("::", ColorsHighlight.purple),
            translate("Installing repository dependencies for {}").format(
                bold_line(", ".join(self.package_names))),
        ))

        # @TODO: add support for --skip-failed-build here:
        retry_interactive_command_or_exit(
            sudo([
                *self._get_pacman_command(),
                "--sync", "--asdeps",
                *self.all_deps_to_install,
            ]),
            pikspect=True,
            conflicts=self.resolved_conflicts,
        )
        PackageDB.discard_local_cache()
        self._local_pkgs_with_build_deps = set(PackageDB.get_local_dict().keys())
        self._local_provided_pkgs_with_build_deps = PackageDB.get_local_provided_dict()

    def install_all_deps(self, all_package_builds: dict[str, "PackageBuild"]) -> None:
        with FileLock(BUILD_DEPS_LOCK):
            self.get_deps(all_package_builds)
            if self.all_deps_to_install or self.built_deps_to_install:
                PackageDB.discard_local_cache()
                self._local_pkgs_wo_build_deps = set(PackageDB.get_local_dict().keys())
            self.install_built_deps(all_package_builds)
            self._install_repo_deps()

    def _remove_installed_deps(self) -> None:
        if not self._local_pkgs_wo_build_deps:
            return

        logger.debug("Gonna compute diff of installed pkgs")
        deps_packages_installed = self._local_pkgs_with_build_deps.difference(
            self._local_pkgs_wo_build_deps,
        )
        deps_packages_removed = self._local_pkgs_wo_build_deps.difference(
            self._local_pkgs_with_build_deps,
        )
        logger.debug("Deps installed: {}", deps_packages_installed)
        logger.debug("Deps removed: {}", deps_packages_removed)
        if not deps_packages_installed:
            return

        # check if there is diff incosistency because of the package replacement:
        if deps_packages_removed:
            for removed_pkg_name in list(deps_packages_removed):
                for installed_pkg_name in list(deps_packages_installed):
                    if (
                            removed_pkg_name in self._local_provided_pkgs_with_build_deps
                    ) and (
                        installed_pkg_name
                        in [
                            dep.name for dep
                            in self._local_provided_pkgs_with_build_deps[removed_pkg_name]
                        ]
                    ):
                        deps_packages_installed.remove(installed_pkg_name)
                        deps_packages_removed.remove(removed_pkg_name)
                        continue

        if deps_packages_removed:
            error_text = translate(
                "Failed to remove installed dependencies, packages inconsistency: {}",
            ).format(
                bold_line(", ".join(deps_packages_removed)),
            )
            print_error(error_text)
            if not ask_to_continue():
                raise DependencyError(error_text)
        if not deps_packages_installed or self.args.keepbuilddeps:
            return

        print_stderr("{} {}:".format(  # pylint: disable=consider-using-f-string
            color_line("::", ColorsHighlight.purple),
            translate("Removing already installed dependencies for {}").format(
                bold_line(", ".join(self.package_names))),
        ))
        retry_interactive_command_or_exit(
            sudo(
                # pacman --remove flag conflicts with some --sync options:
                [
                    *self._get_pacman_command(ignore_args=["overwrite"]),
                    "--remove",
                    *list(deps_packages_installed),
                ],
            ),
            pikspect=True,
        )
        PackageDB.discard_local_cache()

    def check_pkg_arch(self) -> None:
        if self.skip_carch_check:
            return

        src_info = SrcInfo(pkgbuild_path=self.pkgbuild_path)
        arch = MakepkgConfig.get("CARCH")
        supported_archs = src_info.get_values("arch")
        if supported_archs and (
                ARCH_ANY not in supported_archs
        ) and (
            arch not in supported_archs
        ):
            error_text = translate(
                "{name} can't be built on the current arch ({arch}). "
                "Supported: {suparch}",
            ).format(
                name=bold_line(", ".join(self.package_names)),
                arch=arch,
                suparch=", ".join(supported_archs),
            )
            print_error(error_text)
            if (
                    not self.args.skip_failed_build
            ) and (
                not ask_to_continue(default_yes=False)
            ):
                raise BuildError(error_text)
            self.skip_carch_check = True

    def build_with_makepkg(self) -> bool:  # pylint: disable=too-many-branches,too-many-statements
        makepkg_args = []
        if not self.args.needed:
            makepkg_args.append("--force")
        if not color_enabled():
            makepkg_args.append("--nocolor")

        print_stderr("\n{} {}:".format(  # pylint: disable=consider-using-f-string
            color_line("::", ColorsHighlight.purple),
            translate("Starting the build"),
        ))
        build_succeeded = False
        skip_pgp_check = False
        skip_file_checksums = False
        skip_check = False
        no_prepare = False
        while True:
            cmd_args = MakePkgCommand.get() + makepkg_args
            if skip_pgp_check:
                cmd_args += ["--skippgpcheck"]
            if skip_file_checksums:
                cmd_args += ["--skipchecksums"]
            if self.skip_carch_check:
                cmd_args += ["--ignorearch"]
            if skip_check:
                cmd_args += ["--nocheck"]
            if no_prepare:
                cmd_args += ["--noprepare"]

            env = {}
            if self.build_gpgdir:
                env["GNUPGHOME"] = self.build_gpgdir

            cmd_args = isolate_root_cmd(cmd_args, cwd=self.build_dir, env=env)
            spawn_kwargs: "SpawnArgs" = {
                "cwd": str(self.build_dir),
                "env": {**os.environ, **env},
            }
            if self.args.hide_build_log:
                spawn_kwargs.update({
                    "stdout": PIPE,
                    "stderr": PIPE,
                })

            result = interactive_spawn(
                cmd_args,
                **spawn_kwargs,
            )
            print_stdout()
            build_succeeded = result.returncode == 0
            if build_succeeded:
                break

            print_stderr(
                color_line(
                    translate("Command '{}' failed to execute.").format(
                        " ".join(cmd_args),
                    ),
                    ColorsHighlight.red,
                ),
            )
            if self.args.skip_failed_build:
                answer = translate("s")
            elif self.args.noconfirm:
                answer = translate("a")
            else:  # pragma: no cover
                prompt = "{} {}\n{}\n> ".format(
                    color_line("::", ColorsHighlight.yellow),
                    translate("Try recovering?"),
                    "\n".join((
                        translate("[R] retry build"),
                        translate("[p] PGP check skip"),
                        translate("[c] checksums skip"),
                        translate("[f] skip 'check()' function of PKGBUILD"),
                        translate("[n] skip 'prepare()' function of PKGBUILD"),
                        translate("[i] ignore architecture"),
                        translate("[d] delete build dir and try again"),
                        translate("[e] edit PKGBUILD"),
                        "-" * 24,
                        translate("[s] skip building this package"),
                        translate("[a] abort building all the packages"),
                    )),
                )
                answer = get_input(
                    prompt,
                    translate("r").upper() +
                    translate("p") +
                    translate("c") +
                    translate("f") +
                    translate("n") +
                    translate("i") +
                    translate("d") +
                    translate("e") +
                    translate("s") +
                    translate("a"),
                )

            answer = answer.lower()[0]
            if answer == translate("r"):  # pragma: no cover
                continue
            if answer == translate("p"):  # pragma: no cover
                skip_pgp_check = True
                continue
            if answer == translate("c"):  # pragma: no cover
                skip_file_checksums = True
                continue
            if answer == translate("f"):  # pragma: no cover
                skip_check = True
                continue
            if answer == translate("n"):  # pragma: no cover
                no_prepare = True
                continue
            if answer == translate("i"):  # pragma: no cover
                self.skip_carch_check = True
                continue
            if answer == translate("d"):  # pragma: no cover
                self.prepare_build_destination(flush=True)
                continue
            if answer == translate("e"):  # pragma: no cover
                editor_cmd = get_editor_or_exit()
                if editor_cmd:
                    interactive_spawn(
                        [*editor_cmd, str(self.pkgbuild_path)],
                    )
                    interactive_spawn(isolate_root_cmd([
                        "cp",
                        str(self.pkgbuild_path),
                        str(self.build_dir / DEFAULT_PKGBUILD_BASENAME),
                    ]))
                    raise PkgbuildChanged
                continue
            if answer == translate("a"):
                raise SysExit(125)
            # "s"kip
            break

        return build_succeeded

    def build(
            self,
            all_package_builds: dict[str, "PackageBuild"],
            resolved_conflicts: list[list[str]],
    ) -> None:
        self.resolved_conflicts = resolved_conflicts

        self.prepare_build_destination()

        self.install_all_deps(all_package_builds)
        try:
            build_succeeded = True if self.check_if_already_built() else self.build_with_makepkg()
        finally:
            self._remove_installed_deps()

        if not build_succeeded:
            self.failed = True
            raise BuildError
        self._set_built_package_path()


class AlreadyClonedRepos:

    repos: list[str] = []

    @classmethod
    def add(cls, repo: str) -> None:
        cls.repos.append(repo)

    @classmethod
    def get(cls, repo: str) -> bool:
        return repo in cls.repos


def clone_aur_repos(package_names: list[str]) -> dict[str, PackageBuild]:
    aur_pkgs, _ = find_aur_packages(package_names)
    packages_bases: dict[str, list[str]] = {}
    for aur_pkg in aur_pkgs:
        packages_bases.setdefault(aur_pkg.packagebase, []).append(aur_pkg.name)
    package_builds_by_base = {
        pkgbase: PackageBuild(pkg_names)
        for pkgbase, pkg_names in packages_bases.items()
        if not AlreadyClonedRepos.get(pkgbase)
    }
    pool_size: int | None = None
    if clone_c := parse_args().aur_clone_concurrency:
        pool_size = clone_c
    elif running_as_root():
        pool_size = 1
    exc: CloneError | None = None
    with ThreadPool(processes=pool_size) as pool:
        requests = {
            key: pool.apply_async(repo_status.update_aur_repo, ())
            for key, repo_status in package_builds_by_base.items()
        }
        pool.close()
        pool.join()
        for package_base, request in requests.items():
            result = request.get()
            if result and result.returncode > 0:
                exc = CloneError(
                    build=package_builds_by_base[package_base],
                    result=result,
                )
            else:
                AlreadyClonedRepos.add(package_base)
    if exc:
        raise exc
    all_package_builds_by_base = {
        pkgbase: PackageBuild(pkg_names)
        for pkgbase, pkg_names in packages_bases.items()
    }
    return {
        pkg_name: all_package_builds_by_base[pkgbase]
        for pkgbase, pkg_names in packages_bases.items()
        for pkg_name in pkg_names
    }



"""Licensed under GPLv3, see https://www.gnu.org/licenses/"""

import sys
import tty
from typing import TYPE_CHECKING

from .args import LiteralArgs, parse_args
from .config import PROMPT_LOCK, PikaurConfig
from .core import get_editor, interactive_spawn
from .exceptions import SysExit
from .filelock import FileLock
from .i18n import translate
from .logging import create_logger
from .pikspect import ReadlineKeycodes, TTYInputWrapper, TTYRestore
from .pikspect import pikspect as pikspect_spawn
from .pprint import (
    ColorsHighlight,
    color_line,
    get_term_width,
    print_stderr,
    print_warning,
    range_printable,
)

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence
    from typing import Final


logger = create_logger("prompt")
logger_no_lock = create_logger("prompt_nolock", lock=False)


class Answers():

    _init_done: bool = False
    Y: str  # pylint: disable=invalid-name
    N: str  # pylint: disable=invalid-name
    Y_UP: str
    N_UP: str

    @classmethod
    def _do_init(cls) -> None:
        if not cls._init_done:
            cls.Y = translate("y")
            cls.N = translate("n")
            cls.Y_UP = cls.Y.upper()
            cls.N_UP = cls.N.upper()
            cls._init_done = True

    def __init__(self) -> None:
        self._do_init()


def read_answer_from_tty(question: str, answers: "Sequence[str] | None" = None) -> str:
    """
    Function displays a question and reads a single character
    from STDIN as an answer. Then returns the character as lower character.
    Valid answers are passed as 'answers' variable (the default is in capital).
    Invalid answer will return an empty string.
    """
    default = " "
    all_answers = Answers()
    answers = answers or (all_answers.Y_UP, all_answers.N)

    for letter in answers:
        if letter.isupper():
            default = letter.lower()
            break

    if not sys.stdin.isatty():
        return default

    print_stderr(question, flush=True, end=" ", lock=False)
    previous_tty_settings = tty.tcgetattr(sys.stdin.fileno())  # type: ignore[attr-defined]
    try:
        tty.setraw(sys.stdin.fileno())
        answer = sys.stdin.read(1).lower()
    except Exception:
        return " "
    else:
        if ord(answer) in (ReadlineKeycodes.CTRL_C, ReadlineKeycodes.CTRL_D):
            raise SysExit(1)
        if ord(answer) == ReadlineKeycodes.ENTER:
            answer = default
            return default
        if answer in [choice.lower() for choice in answers]:
            return answer
        return " "
    finally:
        tty.tcsetattr(  # type: ignore[attr-defined]
            sys.stdin.fileno(), tty.TCSADRAIN, previous_tty_settings,  # type: ignore[attr-defined]
        )
        sys.stdout.write(f"{answer}\r\n")
        tty.tcdrain(sys.stdin.fileno())  # type: ignore[attr-defined]


def split_last_line(text: str) -> str:
    all_lines = text.split("\n")
    n_lines = len(all_lines)
    last_line = all_lines[n_lines - 1]
    term_width = get_term_width()
    if len(last_line) < term_width:
        return text
    prev_lines = all_lines[:n_lines - 1]
    return "\n".join([
        *prev_lines,
        range_printable(last_line, 0, term_width),
        range_printable(last_line, term_width),
    ])


def get_input(
        prompt: str, answers: "Sequence[str]" = (), *, require_confirm: bool = False,
) -> str:
    logger.debug("Gonna get input from user...")
    answer = ""
    with (
            FileLock(PROMPT_LOCK),
            TTYInputWrapper(),
    ):
        if not (
                require_confirm or PikaurConfig().ui.RequireEnterConfirm.get_bool()
        ):
            logger_no_lock.debug("Using custom input reader...")
            answer = read_answer_from_tty(prompt, answers=answers)
        else:
            logger_no_lock.debug("Restoring TTY...")
            sub_tty = TTYRestore()
            TTYRestore.restore()
            try:
                logger_no_lock.debug("Using standard input reader...")
                answer = input(split_last_line(prompt)).lower()
            except EOFError as exc:
                logger_no_lock.debug(exc)
                raise SysExit(125) from exc
            finally:
                logger_no_lock.debug("Reverting to prev TTY state...")
                sub_tty.restore_new()

    if not answer:
        for choice in answers:
            if choice.isupper():
                logger.debug('No answer provided - using "{}".', choice)
                return choice.lower()
    logger.debug("Got answer: '{}'", answer)
    return answer


class NotANumberInputError(Exception):
    character: str

    def __init__(self, character: str) -> None:
        self.character = character
        super().__init__(f'"{character} is not a number')


class NumberRangeInputSyntax:
    DELIMITERS: "Final[Sequence[str]]" = (" ", ",")
    RANGES: "Final[Sequence[str]]" = ("-", "..")


def get_multiple_numbers_input(prompt: str, answers: "Iterable[int]" = ()) -> list[int]:
    str_result = get_input(prompt, [str(answer) for answer in answers], require_confirm=True)
    if not str_result:
        return []
    for delimiter in NumberRangeInputSyntax.DELIMITERS[1:]:
        str_result = str_result.replace(delimiter, NumberRangeInputSyntax.DELIMITERS[0])
    str_results = str_result.split(NumberRangeInputSyntax.DELIMITERS[0])
    int_results: list[int] = []
    for raw_block in str_results[:]:
        block = raw_block
        for range_char in NumberRangeInputSyntax.RANGES:
            block = block.replace(range_char, NumberRangeInputSyntax.RANGES[0])
        if NumberRangeInputSyntax.RANGES[0] in block:
            try:
                range_start, range_end = (
                    int(char) for char in block.split(NumberRangeInputSyntax.RANGES[0], maxsplit=1)
                )
            except ValueError as exc:
                raise NotANumberInputError(block) from exc
            if range_start > range_end:
                raise NotANumberInputError(block)
            int_results += list(range(range_start, range_end + 1))
        else:
            try:
                int_results.append(int(block))
            except ValueError as exc:
                raise NotANumberInputError(exc.args[0].split("'")[-2]) from exc
    return int_results


def ask_to_continue(text: str | None = None, *, default_yes: bool = True) -> bool:
    args = parse_args()
    if text is None:
        text = translate("Do you want to proceed?")

    if args.noconfirm:
        default_option = translate(
            "[Y]es ({reason})"
            if default_yes else
            "[N]o ({reason})",
        ).format(reason=LiteralArgs.NOCONFIRM)
        print_stderr(f"{text} {default_option}")
        return default_yes

    all_answers = Answers()
    prompt = text + (
        f" [{all_answers.Y_UP}/{all_answers.N}] " if default_yes else
        f" [{all_answers.Y}/{all_answers.N_UP}] "
    )
    answers = (
        (all_answers.Y_UP + all_answers.N) if default_yes else
        (all_answers.Y + all_answers.N_UP)
    )

    answer = get_input(prompt, answers)
    return (answer == all_answers.Y) or (default_yes and not answer)


def retry_interactive_command(
        cmd_args: list[str],
        *,
        pikspect: bool = False,
        conflicts: list[list[str]] | None = None,
) -> bool:
    args = parse_args()
    while True:
        good = (
            pikspect_spawn(
                cmd_args,
                conflicts=conflicts,
            )
            if pikspect and (LiteralArgs.NOCONFIRM not in cmd_args) else
            interactive_spawn(
                cmd_args,
            )
        ).returncode == 0
        if good:
            return good
        print_stderr(color_line(
            translate("Command '{}' failed to execute.").format(
                " ".join(cmd_args),
            ),
            ColorsHighlight.red,
        ))
        if not ask_to_continue(
                text=translate("Do you want to retry?"),
                default_yes=not args.noconfirm,
        ):
            return False


def retry_interactive_command_or_exit(
        cmd_args: list[str],
        *,
        pikspect: bool = False,
        conflicts: list[list[str]] | None = None,
) -> None:
    if not retry_interactive_command(
            cmd_args,
            pikspect=pikspect,
            conflicts=conflicts,
    ) and not ask_to_continue(default_yes=False):
        raise SysExit(125)


def get_editor_or_exit() -> list[str] | None:
    editor = get_editor()
    if not editor:
        print_warning(translate("no editor found. Try setting $VISUAL or $EDITOR."))
        if not ask_to_continue(
                translate("Do you want to proceed without editing?"),
        ):  # pragma: no cover
            raise SysExit(125)
    return editor



"""Licensed under GPLv3, see https://www.gnu.org/licenses/"""

# pylint: disable=too-many-lines
import contextlib
import hashlib
from multiprocessing.pool import ThreadPool
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import TYPE_CHECKING

from .args import parse_args, reconstruct_args
from .aur import AURPackageInfo
from .build import PackageBuild, PkgbuildChanged, clone_aur_repos
from .config import DiffPagerValues, PikaurConfig
from .conflicts import find_aur_conflicts
from .core import (
    PackageSource,
    interactive_spawn,
    isolate_root_cmd,
    open_file,
    remove_dir,
    running_as_root,
    sudo,
)
from .exceptions import (
    BuildError,
    CloneError,
    DependencyError,
    DependencyNotBuiltYetError,
    DependencyVersionMismatchError,
    PackagesNotFoundInAURError,
    SysExit,
)
from .i18n import translate
from .install_info_fetcher import InstallInfoFetcher
from .logging import create_logger
from .news import News
from .pacman import (
    PackageDB,
    get_pacman_command,
    install_built_deps,
    refresh_pkg_db_if_needed,
    strip_repo_name,
)
from .pikspect import TTYRestore
from .pprint import (
    ColorsHighlight,
    bold_line,
    color_line,
    print_error,
    print_stderr,
    print_stdout,
    print_warning,
)
from .print_department import (
    pretty_format_sysupgrade,
    print_local_package_newer,
    print_not_found_packages,
    print_package_downgrading,
    print_package_uptodate,
)
from .prompt import (
    ask_to_continue,
    get_editor_or_exit,
    get_input,
    retry_interactive_command,
    retry_interactive_command_or_exit,
)
from .srcinfo import SrcInfo
from .updates import is_devel_pkg
from .version import compare_versions

if TYPE_CHECKING:
    import pyalpm

    from .args import PikaurArgs

logger = create_logger("install_cli")


def hash_file(filename: str | Path) -> str:  # pragma: no cover
    filename = Path(filename)
    md5 = hashlib.new("md5", usedforsecurity=False)
    with filename.open("rb") as file:
        eof = False
        while not eof:
            data = file.read(1024)
            if data:
                md5.update(data)
            else:
                eof = True
    return md5.hexdigest()


def edit_file(filename: str | Path) -> bool:  # pragma: no cover
    editor_cmd = get_editor_or_exit()
    if not editor_cmd:
        return False
    old_hash = hash_file(filename)
    sub_tty = TTYRestore()
    TTYRestore.restore()
    interactive_spawn([
        *editor_cmd, str(filename),
    ])
    sub_tty.restore_new()
    new_hash = hash_file(filename)
    return old_hash != new_hash


class InstallPackagesCLI():

    # User input
    args: "PikaurArgs"
    install_package_names: list[str]
    # @TODO: define @property for manually_excluded_packages_names+args.ignore:
    manually_excluded_packages_names: list[str]
    resolved_conflicts: list[list[str]]
    reviewed_package_bases: list[str]
    # pkgbuild_path: [pkg_name, ...]  -- needed for split pkgs to install only some of them
    pkgbuilds_packagelists: dict[str, list[str]]

    # computed package lists:
    not_found_repo_pkgs_names: list[str]
    found_conflicts: dict[str, list[str]]
    repo_packages_by_name: "dict[str, pyalpm.Package]"
    # pkgbuilds from cloned aur repos:
    package_builds_by_name: dict[str, PackageBuild]

    # Packages' install info
    install_info: InstallInfoFetcher

    # Installation results
    # transactions by PackageSource(AUR/repo), direction(removed/installed):
    transactions: dict[str, dict[str, list[str]]]
    # AUR packages which failed to build:
    # @TODO: refactor to store in transactions
    failed_to_build_package_names: list[str]

    # arch news
    news: News | None = None

    def __init__(self) -> None:
        self.args = parse_args()
        self.install_package_names = self.args.positional[:]

        self.pkgbuilds_packagelists = {}
        self.manually_excluded_packages_names = []
        self.resolved_conflicts = []
        self.reviewed_package_bases = []

        self.not_found_repo_pkgs_names = []
        self.repo_packages_by_name = {}
        self.package_builds_by_name = {}

        self.found_conflicts = {}
        self.transactions = {}
        self.failed_to_build_package_names = []

        if not self.args.aur and (self.args.sysupgrade or self.args.refresh):

            with ThreadPool() as pool:
                threads = []
                if self.args.sysupgrade:
                    self.news = News()
                    threads.append(
                        pool.apply_async(self.news.fetch_latest, ()),
                    )
                if self.args.refresh:
                    threads.append(
                        pool.apply_async(refresh_pkg_db_if_needed, ()),
                    )
                pool.close()
                for thread in threads:
                    thread.get()
                pool.join()

            if not (self.install_package_names or self.args.sysupgrade):
                return

            if self.args.refresh:
                PackageDB.discard_repo_cache()
                print_stdout()

        if self.args.sysupgrade and not self.args.repo:
            print_stderr("{} {}".format(  # pylint: disable=consider-using-f-string
                color_line("::", ColorsHighlight.blue),
                bold_line(translate("Starting full AUR upgrade...")),
            ))
        if self.args.aur:
            self.not_found_repo_pkgs_names = self.install_package_names
            self.install_package_names = []
        if self.args.pkgbuild:
            self.get_info_from_pkgbuilds()

        self.main_sequence()

    class ExitMainSequence(Exception):  # noqa: N818
        """Raise when need to finish Install CLI"""

    def main_sequence(self) -> None:
        with contextlib.suppress(self.ExitMainSequence):
            self.get_all_packages_info()
            if self.news:
                self.news.print_news()
            if not self.args.noconfirm:
                self.install_prompt()

            self.get_package_builds()
            # @TODO: ask to install optdepends (?)
            if not self.args.downloadonly:
                self.ask_about_package_conflicts()
            self.review_build_files()

            self.install_packages()

    @property
    def aur_packages_names(self) -> list[str]:
        return self.install_info.aur_packages_names

    @property
    def aur_deps_names(self) -> list[str]:
        return self.install_info.aur_deps_names

    @property
    def all_aur_packages_names(self) -> list[str]:
        return list(set(self.aur_packages_names + self.aur_deps_names))

    def get_info_from_pkgbuilds(self) -> None:
        self.install_package_names = []
        self.not_found_repo_pkgs_names = []
        self.pkgbuilds_packagelists = {
            path: [] for path in
            self.args.positional or ["PKGBUILD"]
        }

    def aur_pkg_not_found_prompt(self, pkg_name: str) -> None:  # pragma: no cover
        prompt = "{} {}\n{}\n{}\n{}\n> ".format(  # pylint: disable=consider-using-f-string
            color_line("::", ColorsHighlight.yellow),
            translate("Try recovering {pkg_name}?").format(pkg_name=bold_line(pkg_name)),
            translate("[e] edit PKGBUILD"),
            translate("[s] skip this package"),
            translate("[A] abort"),
        )
        answer = get_input(prompt, translate("e") + translate("s") + translate("a").upper())

        answer = answer.lower()[0]
        if answer == translate("e"):
            updated_pkgbuilds = self._clone_aur_repos([pkg_name])
            if not updated_pkgbuilds:
                return
            self.package_builds_by_name.update(updated_pkgbuilds)
            pkg_build = self.package_builds_by_name[pkg_name]
            if not edit_file(
                    pkg_build.pkgbuild_path,
            ):
                print_warning(translate("PKGBUILD appears unchanged after editing"))
            else:
                self.handle_pkgbuild_changed(pkg_build)
            self._ignore_package(pkg_name)
            self.pkgbuilds_packagelists[str(pkg_build.pkgbuild_path)] = pkg_build.package_names
            self.main_sequence()
        elif answer == translate("s"):
            self._ignore_package(pkg_name)
        else:
            raise SysExit(125)

    def get_all_packages_info(self) -> None:  # pylint:disable=too-many-branches,too-many-statements
        """
        Retrieve info (`InstallInfo` objects) of packages
        which are going to be installed/upgraded and their dependencies
        """
        # deal with package names which user explicitly wants to install
        self.repo_packages_by_name = {}

        for pkg_name in self.manually_excluded_packages_names:
            if pkg_name in self.install_package_names:
                self.install_package_names.remove(pkg_name)

        try:
            self.install_info = InstallInfoFetcher(
                install_package_names=self.install_package_names,
                not_found_repo_pkgs_names=self.not_found_repo_pkgs_names,
                pkgbuilds_packagelists=self.pkgbuilds_packagelists,
                manually_excluded_packages_names=(
                    self.manually_excluded_packages_names + self.args.ignore
                ),
            )
        except PackagesNotFoundInAURError as exc:
            if exc.wanted_by:
                print_error(bold_line(
                    translate("Dependencies missing for {}").format(", ".join(exc.wanted_by)),
                ))
                print_not_found_packages(exc.packages)
                for pkg_name in exc.wanted_by:  # pylint: disable=not-an-iterable
                    self.aur_pkg_not_found_prompt(pkg_name)
                self.get_all_packages_info()
                return
            print_not_found_packages(exc.packages)
            raise SysExit(131) from exc
        except DependencyVersionMismatchError as exc:
            print_stderr(color_line(translate("Version mismatch:"), ColorsHighlight.yellow))
            print_stderr(
                translate("{what} depends on: '{dep}'\n found in '{location}': '{version}'").format(
                    what=bold_line(exc.who_depends),
                    dep=exc.dependency_line,
                    location=exc.location,
                    version=exc.version_found,
                ),
            )
            raise SysExit(131) from exc
        except DependencyError as exc:
            print_stderr(str(exc))
            raise SysExit(131) from exc

        if self.args.repo and self.not_found_repo_pkgs_names:
            print_not_found_packages(self.not_found_repo_pkgs_names, repo=True)
            raise SysExit(6)

        if self.args.needed:
            # check if there are really any new packages need to be installed
            need_refetch_info = False
            logger.debug("checking for --needed")
            logger.debug("before:")
            logger.debug(
                "self.install_info.all_install_info_containers={}",
                self.install_info.all_install_info_containers,
            )
            for install_info in self.install_info.all_install_info:
                pkg_name = install_info.name
                if (
                        is_devel_pkg(pkg_name) and
                        (install_info in self.install_info.aur_updates_install_info)
                ):
                    logger.debug(
                        "'{}' is devel - check it later after retrieving the sources", pkg_name,
                    )
                    continue
                if (
                    not install_info.current_version
                ):
                    logger.debug(
                        "'{}' is not installed", pkg_name,
                    )
                    continue
                if compare_versions(
                    install_info.current_version,
                    install_info.new_version,
                ):
                    logger.debug(
                        "'{}' is need upgrade", pkg_name,
                    )
                    continue
                if (
                    pkg_name not in self.install_package_names
                ):
                    logger.debug(
                        "'{}' package installed via Provides, not by its real name", pkg_name,
                    )
                    continue
                print_package_uptodate(pkg_name, install_info.package_source)
                self.discard_install_info(pkg_name)
                need_refetch_info = True
            if need_refetch_info:
                self.get_all_packages_info()
                return
            logger.debug("after:")

        logger.debug(
            "self.install_info.all_install_info_containers={}",
            self.install_info.all_install_info_containers,
        )
        # check if we really need to build/install anything
        if not self.install_info.all_install_info:
            if not self.args.aur and self.args.sysupgrade:
                self.install_repo_packages()
            else:
                print_stdout(" ".join((
                    color_line("::", ColorsHighlight.green),
                    translate("Nothing to do."),
                )))
            raise SysExit(0)

    def _ignore_package(self, pkg_name: str) -> None:
        self.manually_excluded_packages_names.append(pkg_name)
        for name in (pkg_name, strip_repo_name(pkg_name)):
            if name in self.install_package_names:
                self.install_package_names.remove(name)
            if name in self.not_found_repo_pkgs_names:
                self.not_found_repo_pkgs_names.remove(name)

    def manual_package_selection(self) -> None:  # pragma: no cover

        def parse_pkg_names(text: str) -> set[str]:
            selected_packages = []
            for raw_line in text.splitlines():
                line = raw_line.lstrip()
                if not line:
                    continue
                if not line.startswith("::") and not line.startswith("#"):
                    pkg_name = line.split()[0]
                    # for provided package selection: (mb later for optional deps)
                    pkg_name = pkg_name.split("#")[0].strip()
                    selected_packages.append(pkg_name)
            return set(selected_packages)

        text_before = pretty_format_sysupgrade(
            install_info=self.install_info,
            manual_package_selection=True,
        )
        pkg_names_before = parse_pkg_names(text_before)
        with NamedTemporaryFile() as tmp_file:
            with open_file(tmp_file.name, "w") as write_file:
                write_file.write(text_before)
            edit_file(tmp_file.name)
            with open_file(tmp_file.name, "r") as read_file:
                selected_packages = parse_pkg_names(read_file.read())

        list_diff = selected_packages.difference(pkg_names_before)
        for pkg_name in list_diff:
            if pkg_name not in (
                    self.install_package_names + self.not_found_repo_pkgs_names
            ):
                self.install_package_names.append(pkg_name)

        for pkg_name in pkg_names_before.difference(selected_packages):
            self._ignore_package(pkg_name)

    def install_prompt(self) -> None:  # pragma: no cover

        def _print_sysupgrade(*, verbose: bool = False) -> None:
            print_stdout(pretty_format_sysupgrade(
                install_info=self.install_info,
                verbose=verbose,
            ))

        def _confirm_sysupgrade(*, verbose: bool = False) -> str:
            _print_sysupgrade(verbose=verbose)
            prompt = "{} {}\n{} {}\n>> ".format(  # pylint: disable=consider-using-f-string
                color_line("::", ColorsHighlight.blue),
                bold_line(translate("Proceed with installation? [Y/n] ")),
                color_line("::", ColorsHighlight.blue),
                bold_line(translate("[v]iew package details   [m]anually select packages")),
            )
            return get_input(
                prompt,
                translate("y").upper() + translate("n") + translate("v") + translate("m"),
            )

        if self.args.noconfirm:
            _print_sysupgrade()
            return
        answer = None
        while True:
            if answer is None:
                answer = _confirm_sysupgrade()
            if answer:
                letter = answer.lower()[0]
                if letter == translate("y"):
                    break
                if letter == translate("v"):
                    answer = _confirm_sysupgrade(verbose=True)
                    continue
                if letter == translate("m"):
                    print_stdout()
                    self.manual_package_selection()
                    self.get_all_packages_info()
                    self.install_prompt()
                    break
                raise SysExit(125)
            break

    def discard_install_info(self, canceled_pkg_name: str, *, ignore: bool = True) -> None:
        logger.debug("discarding install info for pkg '{}'...", canceled_pkg_name)
        if ignore:
            logger.debug("ignoring pkg '{}'...", canceled_pkg_name)
            self.manually_excluded_packages_names.append(canceled_pkg_name)
        if not getattr(self, "install_info", None):  # @TODO: make it nicer?
            logger.debug("install info not initialized yet -- running on early stage?")
            return
        for pkg_name in self.install_info.discard_package(canceled_pkg_name):
            logger.debug("discarded install info for pkg: {}", pkg_name)
            if pkg_name in self.install_package_names:
                self.install_package_names.remove(pkg_name)
            if pkg_name in self.not_found_repo_pkgs_names:
                self.not_found_repo_pkgs_names.remove(pkg_name)
            if pkg_name in self.package_builds_by_name:
                del self.package_builds_by_name[pkg_name]

    def _find_extra_aur_build_deps(self, all_package_builds: dict[str, PackageBuild]) -> None:
        need_to_show_install_prompt = False
        for pkgbuild in all_package_builds.values():
            pkgbuild.get_deps(
                all_package_builds=all_package_builds,
                filter_built=False,
                exclude_pkg_names=self.manually_excluded_packages_names,
            )

            aur_pkgs: list[AURPackageInfo] = [
                info.package
                for info in self.install_info.aur_install_info
                if info.name in pkgbuild.package_names
            ]
            aur_rpc_deps = {
                dep_line
                for pkg in aur_pkgs
                for matcher in (
                    pkg.depends +
                    pkg.makedepends +
                    pkg.checkdepends
                )
                for dep_line in matcher.split(",")
            }

            srcinfo_deps: set[str] = set()
            for package_name in pkgbuild.package_names:
                if package_name in self.manually_excluded_packages_names:
                    continue
                src_info = SrcInfo(pkgbuild_path=pkgbuild.pkgbuild_path, package_name=package_name)
                srcinfo_deps.update({
                    dep_line
                    for matcher in
                    list(src_info.get_depends().values()) +
                    list(src_info.get_build_makedepends().values()) +
                    list(src_info.get_build_checkdepends().values())
                    for dep_line in matcher.line.split(",")
                })

            if aur_rpc_deps != srcinfo_deps:
                deps_added = srcinfo_deps.difference(aur_rpc_deps)
                deps_removed = aur_rpc_deps.difference(srcinfo_deps)
                if deps_added:
                    print_warning(
                        translate("New build deps found for {pkg} package: {deps}").format(
                            pkg=bold_line(", ".join(pkgbuild.package_names)),
                            deps=bold_line(", ".join(deps_added)),
                        ),
                    )
                if deps_removed:
                    print_warning(
                        translate("Some build deps removed for {pkg} package: {deps}").format(
                            pkg=bold_line(", ".join(pkgbuild.package_names)),
                            deps=bold_line(", ".join(deps_removed)),
                        ),
                    )
                for pkg_name in pkgbuild.package_names:
                    self.discard_install_info(pkg_name, ignore=False)
                self.pkgbuilds_packagelists[str(pkgbuild.pkgbuild_path)] = pkgbuild.package_names
                need_to_show_install_prompt = True
        if need_to_show_install_prompt:
            self.main_sequence()
            raise self.ExitMainSequence

    def _clone_aur_repos(  # pylint: disable=too-many-branches
            self, package_names: list[str],
    ) -> dict[str, PackageBuild] | None:
        stash_pop_list: list[str] = []
        while True:
            try:
                pkgbuild_by_name = clone_aur_repos(package_names=package_names)
            except CloneError as err:
                package_build = err.build
                print_stderr(color_line(
                    (
                        translate("Can't clone '{name}' in '{path}' from AUR:")
                        if package_build.clone else
                        translate("Can't pull '{name}' in '{path}' from AUR:")
                    ).format(
                        name=", ".join(package_build.package_names),
                        path=package_build.repo_path,
                    ),
                    ColorsHighlight.red,
                ))
                print_stderr(err.result.stdout_text)
                print_stderr(err.result.stderr_text)
                if self.args.noconfirm:
                    answer = translate("a")
                else:  # pragma: no cover
                    prompt = "{} {}\n> ".format(
                        color_line("::", ColorsHighlight.yellow),
                        "\n".join((
                            translate("Try recovering?"),
                            translate("[T] try again"),
                            translate("[c] git checkout -- '*'"),
                            # translate("[c] git checkout -- '*' ; git clean -f -d -x"),
                            translate("[r] remove dir and clone again"),
                            translate("[p] git stash && ... && git stash pop"),
                            translate("[s] skip this package"),
                            translate("[a] abort"),
                        )),
                    )
                    answer = get_input(
                        prompt,
                        translate("t").upper() + translate("c") + translate("r")
                        + translate("s") + translate("a"),
                    )

                answer = answer.lower()[0]
                if answer == translate("c"):  # pragma: no cover
                    package_build.git_reset_changed()
                elif answer == translate("p"):  # pragma: no cover
                    package_build.git_stash()
                    stash_pop_list.append(package_build.package_base)
                elif answer == translate("r"):  # pragma: no cover
                    remove_dir(package_build.repo_path)
                elif answer == translate("s"):  # pragma: no cover
                    for skip_pkg_name in package_build.package_names:
                        self.discard_install_info(skip_pkg_name)
                        if skip_pkg_name in package_names:
                            package_names.remove(skip_pkg_name)
                elif answer == translate("a"):  # pragma: no cover
                    raise SysExit(125) from err
            else:
                for pkg_build in pkgbuild_by_name.values():
                    if pkg_build.package_base in stash_pop_list:
                        pkg_build.git_stash_pop()
                return pkgbuild_by_name

    def get_package_builds(self) -> None:
        while self.all_aur_packages_names:
            clone_names = []
            pkgbuilds_by_base: dict[str, PackageBuild] = {}
            pkgbuilds_by_name = {}
            for info in self.install_info.aur_install_info:
                if info.pkgbuild_path:
                    if not isinstance(info.package, AURPackageInfo):
                        raise TypeError
                    pkg_base = info.package.packagebase
                    if pkg_base not in pkgbuilds_by_base:
                        package_names = self.pkgbuilds_packagelists.get(info.pkgbuild_path)
                        logger.debug(
                            "Initializing build info for {}({}): {}",
                            pkg_base, package_names, info.pkgbuild_path,
                        )
                        pkgbuilds_by_base[pkg_base] = PackageBuild(
                            pkgbuild_path=info.pkgbuild_path,
                            package_names=package_names,
                        )
                    pkgbuilds_by_name[info.name] = pkgbuilds_by_base[pkg_base]
                else:
                    clone_names.append(info.name)
            cloned_pkgbuilds = self._clone_aur_repos(clone_names)
            if cloned_pkgbuilds:
                pkgbuilds_by_name.update(cloned_pkgbuilds)
            for pkg_list in (self.aur_packages_names, self.aur_deps_names):
                self._find_extra_aur_build_deps(
                    all_package_builds={
                        pkg_name: pkgbuild for pkg_name, pkgbuild
                        in pkgbuilds_by_name.items()
                        if pkg_name in pkg_list
                    },
                )
            self.package_builds_by_name = pkgbuilds_by_name
            break

    def ask_about_package_conflicts(self) -> None:
        if self.aur_packages_names or self.aur_deps_names:
            print_stderr(translate("looking for conflicting AUR packages..."))
            self.found_conflicts.update(
                find_aur_conflicts(
                    self.install_info.aur_install_info,
                    self.install_package_names,
                ),
            )
        if not self.found_conflicts:
            return
        all_new_packages_names = list(self.repo_packages_by_name.keys()) + self.aur_packages_names
        for new_pkg_name, new_pkg_conflicts in self.found_conflicts.items():
            for pkg_conflict in new_pkg_conflicts:
                if pkg_conflict in all_new_packages_names:
                    print_stderr(color_line(
                        translate("New packages '{new}' and '{other}' are in conflict.").format(
                            new=new_pkg_name, other=pkg_conflict),
                        ColorsHighlight.red))
                    raise SysExit(131)
        for new_pkg_name, new_pkg_conflicts in self.found_conflicts.items():
            for pkg_conflict in new_pkg_conflicts:
                answer = ask_to_continue("{} {}".format(  # pylint: disable=consider-using-f-string
                    color_line("::", ColorsHighlight.yellow),
                    bold_line(translate(
                        "{new} and {installed} are in conflict. Remove {installed}?",
                    ).format(
                        new=new_pkg_name, installed=pkg_conflict,
                    )),
                ), default_yes=False)
                if not answer:
                    raise SysExit(131)
                self.resolved_conflicts.append([new_pkg_name, pkg_conflict])

    def ask_to_edit_file(
            self, filename: str, package_build: PackageBuild,
    ) -> bool:  # pragma: no cover
        noedit = not self.args.edit and (
            self.args.noedit
        )
        if noedit or self.args.noconfirm:
            print_stderr("{} {}".format(  # pylint: disable=consider-using-f-string
                color_line("::", ColorsHighlight.yellow),
                translate("Skipping review of {file} for {name} package ({flag})").format(
                    file=filename,
                    name=", ".join(package_build.package_names),
                    flag=(noedit and "--noedit") or
                    (self.args.noconfirm and "--noconfirm")),
            ))
            return False
        if not ask_to_continue(
                translate("Do you want to {edit} {file} for {name} package?").format(
                    edit=bold_line(translate("edit")),
                    file=filename,
                    name=bold_line(", ".join(package_build.package_names)),
                ),
                default_yes=not (package_build.last_installed_hash or
                                 PikaurConfig().review.DontEditByDefault.get_bool()),
        ):
            return False
        full_filename = package_build.repo_path / filename
        return edit_file(full_filename)

    def _get_installed_status(self) -> None:
        all_package_builds = set(self.package_builds_by_name.values())

        # if running as root get sources for dev packages synchronously
        # (to prevent race condition in systemd dynamic users)
        num_threads: int | None = None
        if running_as_root():  # pragma: no cover
            num_threads = 1

        # check if pkgs versions already installed
        # (use threads because devel packages require downloading
        # latest sources for quite a long time)
        with ThreadPool(processes=num_threads) as pool:
            threads = []
            for pkg_build in all_package_builds:
                threads.append(
                    pool.apply_async(getattr, (pkg_build, "version_already_installed")),
                )
            for thread in threads:
                thread.get()
            pool.close()
            pool.join()

        # handle if version is already installed
        if not self.args.needed:
            return
        local_db = PackageDB.get_local_dict()
        for pkg_build in all_package_builds:
            if not pkg_build.reviewed:
                continue
            # pragma: no cover
            pkg_build.update_last_installed_file()
            for package_name in pkg_build.package_names:
                if package_name not in local_db:
                    continue
                if pkg_build.version_already_installed:
                    print_package_uptodate(package_name, PackageSource.AUR)
                    self.discard_install_info(package_name)
                elif (
                    (
                        self.args.sysupgrade > 1
                    ) or (
                        is_devel_pkg(pkg_build.package_base) and (self.args.devel > 1)
                    ) and not pkg_build.version_is_upgradeable
                ):
                    print_package_downgrading(
                        package_name,
                        downgrade_version=pkg_build.get_version(package_name),
                    )
                elif not pkg_build.version_is_upgradeable:
                    print_local_package_newer(
                        package_name,
                        aur_version=pkg_build.get_version(package_name),
                    )
                    self.discard_install_info(package_name)

    def review_build_files(self) -> None:  # pragma: no cover  pylint:disable=too-many-branches
        if self.args.needed or self.args.devel:
            self._get_installed_status()
        for pkg_build in set(self.package_builds_by_name.values()):
            _pkg_label = bold_line(", ".join(pkg_build.package_names))
            _skip_diff_label = translate("Not showing diff for {pkg} package ({reason})")

            if (
                    pkg_build.package_base in self.reviewed_package_bases
            ):
                print_warning(_skip_diff_label.format(
                    pkg=_pkg_label,
                    reason=translate("already reviewed"),
                ))
                continue

            if (
                    pkg_build.last_installed_hash != pkg_build.current_hash
            ) and (
                pkg_build.last_installed_hash
            ) and (
                pkg_build.current_hash
            ) and (
                not self.args.noconfirm
            ) and (
                not self.args.nodiff
            ) and ask_to_continue(
                    translate(
                        "Do you want to see build files {diff} for {name} package?",
                    ).format(
                        diff=bold_line(translate("diff")),
                        name=_pkg_label,
                    ),
            ):
                git_args: list[str] = []
                diff_pager = PikaurConfig().review.DiffPager
                if diff_pager == DiffPagerValues.ALWAYS:
                    git_args = ["env", "GIT_PAGER=less -+F"]
                elif diff_pager == DiffPagerValues.NEVER:
                    git_args = ["env", "GIT_PAGER=cat"]
                git_args += [
                    "git",
                    "-C", str(pkg_build.repo_path),
                    "diff",
                    *PikaurConfig().review.GitDiffArgs.get_str().split(","),
                    pkg_build.last_installed_hash,
                    pkg_build.current_hash,
                    "--", ".",
                ]
                for file_path in PikaurConfig().review.HideDiffFiles.get_str().split(","):
                    if file_path:
                        git_args += [
                            f":(exclude){file_path}",
                        ]
                interactive_spawn(isolate_root_cmd(git_args))
            elif self.args.noconfirm:
                print_stdout(_skip_diff_label.format(
                    pkg=_pkg_label,
                    reason="--noconfirm",
                ))
            elif self.args.nodiff:
                print_stdout(_skip_diff_label.format(
                    pkg=_pkg_label,
                    reason="--nodiff",
                ))
            elif not pkg_build.last_installed_hash:
                print_warning(_skip_diff_label.format(
                    pkg=_pkg_label,
                    reason=translate("installing for the first time"),
                ))
            else:
                print_warning(_skip_diff_label.format(
                    pkg=_pkg_label,
                    reason=translate("already reviewed"),
                ))

            if self.ask_to_edit_file(
                    pkg_build.pkgbuild_path.name, pkg_build,
            ):
                self.handle_pkgbuild_changed(pkg_build)

            for pkg_name in pkg_build.package_names:
                install_src_info = SrcInfo(
                    pkgbuild_path=pkg_build.pkgbuild_path,
                    package_name=pkg_name,
                )
                install_file_name = install_src_info.get_install_script()
                if install_file_name:
                    self.ask_to_edit_file(install_file_name, pkg_build)

            pkg_build.check_pkg_arch()
            pkg_build.reviewed = True
            self.reviewed_package_bases.append(pkg_build.package_base)

    def handle_pkgbuild_changed(self, pkg_build: PackageBuild) -> None:
        logger.debug("handle pkgbuild changed {}", pkg_build)
        for pkg_name in pkg_build.package_names:
            self.discard_install_info(pkg_name, ignore=False)
        src_info = SrcInfo(pkgbuild_path=pkg_build.pkgbuild_path)
        old_srcinfo_hash = hash_file(src_info.path)
        src_info.regenerate()
        new_srcinfo_hash = hash_file(src_info.path)

        self.pkgbuilds_packagelists[str(pkg_build.pkgbuild_path)] = pkg_build.package_names
        self.reviewed_package_bases.append(pkg_build.package_base)

        if not getattr(self, "install_info", None):  # @TODO: make it nicer?
            logger.debug("install info not initialized yet -- running on early stage?")
            if old_srcinfo_hash != new_srcinfo_hash:
                print_warning(translate(
                    "Installation info changed (or new deps found) for {pkg} package",
                ).format(
                    pkg=bold_line(", ".join(pkg_build.package_names)),
                ))
                self.main_sequence()
                raise self.ExitMainSequence
            return

        old_install_info = self.install_info
        self.get_all_packages_info()
        old_install_info.pkgbuilds_packagelists = self.install_info.pkgbuilds_packagelists
        if (
                old_install_info != self.install_info or
                old_srcinfo_hash != new_srcinfo_hash
        ):
            print_warning(translate(
                "Installation info changed (or new deps found) for {pkg} package",
            ).format(
                pkg=bold_line(", ".join(pkg_build.package_names)),
            ))
            self.main_sequence()
            raise self.ExitMainSequence

    def build_packages(self) -> None:  # pylint: disable=too-many-branches
        if self.args.needed or self.args.devel:
            self._get_installed_status()

        failed_to_build_package_names = []
        deps_fails_counter: dict[str, int] = {}
        packages_to_be_built = self.all_aur_packages_names[:]
        index = 0
        while packages_to_be_built:
            logger.debug("Gonna build PKGBUILDS: {}", self.package_builds_by_name)
            if index >= len(packages_to_be_built):
                index = 0

            pkg_name = packages_to_be_built[index]
            pkg_build = self.package_builds_by_name[pkg_name]
            if self.args.needed and pkg_build.version_already_installed:
                packages_to_be_built.remove(pkg_name)
                continue

            try:
                logger.debug("Gonna build pkgnames: {}", pkg_build.package_names)
                pkg_build.build(
                    all_package_builds=self.package_builds_by_name,
                    resolved_conflicts=self.resolved_conflicts,
                )
            except PkgbuildChanged:
                self.handle_pkgbuild_changed(pkg_build)
            except (BuildError, DependencyError) as exc:
                print_stderr(exc)
                print_stderr(
                    color_line(
                        translate("Can't build '{name}'.").format(name=pkg_name) + "\n",
                        ColorsHighlight.red,
                    ),
                )
                # if not ask_to_continue():
                #     raise SysExit(125)
                for _pkg_name in pkg_build.package_names:
                    failed_to_build_package_names.append(_pkg_name)
                    if _pkg_name in packages_to_be_built:
                        packages_to_be_built.remove(_pkg_name)
                    self.discard_install_info(_pkg_name)
                    for remaining_aur_pkg_name in packages_to_be_built[:]:
                        if remaining_aur_pkg_name not in self.all_aur_packages_names:
                            packages_to_be_built.remove(remaining_aur_pkg_name)
            except DependencyNotBuiltYetError as exc:
                index += 1
                for _pkg_name in pkg_build.package_names:
                    deps_fails_counter.setdefault(_pkg_name, 0)
                    deps_fails_counter[_pkg_name] += 1
                    if deps_fails_counter[_pkg_name] > len(self.all_aur_packages_names):
                        print_error(
                            translate(
                                "Dependency cycle detected between {}",
                            ).format(deps_fails_counter),
                        )
                        raise SysExit(131) from exc
            else:
                logger.debug(
                    "Build done for packages {}, removing from queue",
                    pkg_build.package_names,
                )
                for _pkg_name in pkg_build.package_names:
                    if _pkg_name not in self.manually_excluded_packages_names:
                        packages_to_be_built.remove(_pkg_name)

        self.failed_to_build_package_names = failed_to_build_package_names

    def _remove_packages(self, packages_to_be_removed: list[str]) -> None:
        if packages_to_be_removed:
            retry_interactive_command_or_exit(
                sudo([
                    *get_pacman_command(), "-Rs", *packages_to_be_removed,
                ]),
                pikspect=True,
            )
            PackageDB.discard_local_cache()

    def _save_transaction(
            self,
            target: PackageSource,
            removed: list[str] | None = None,
            installed: list[str] | None = None,
    ) -> None:
        target_transaction = self.transactions.setdefault(str(target), {})
        if removed:
            for pkg_name in removed:
                target_transaction.setdefault("removed", []).append(pkg_name)
        if installed:
            for pkg_name in installed:
                target_transaction.setdefault("installed", []).append(pkg_name)

    def _revert_transaction(self, target: PackageSource) -> None:
        if not self.transactions:
            return
        target_transaction = self.transactions.get(str(target))
        if not target_transaction:
            return
        print_warning(
            translate("Reverting {target} transaction...").format(target=target),
        )
        removed = target_transaction.get("removed")
        installed = target_transaction.get("installed")
        if removed:
            pass  # install back
        if installed:
            self._remove_packages(installed)

    def install_repo_packages(self) -> None:
        print_stdout()
        extra_args = []
        if not (self.install_package_names or self.args.sysupgrade):
            return
        for excluded_pkg_name in self.manually_excluded_packages_names + self.args.ignore:
            extra_args.append("--ignore")
            # pacman's --ignore doesn't work with repo name:
            extra_args.append(strip_repo_name(excluded_pkg_name))
        if not retry_interactive_command(
                sudo([
                    *get_pacman_command(),
                    "--sync",
                    *reconstruct_args(self.args, ignore_args=["sync", "ignore", "refresh"]),
                    *self.install_package_names,
                    *extra_args,
                ]),
                pikspect=True,
                conflicts=self.resolved_conflicts,
        ) and not ask_to_continue(default_yes=False):  # pragma: no cover
            self._revert_transaction(PackageSource.REPO)
            raise SysExit(125)
        PackageDB.discard_local_cache()
        self._save_transaction(
            PackageSource.REPO, installed=self.install_package_names,
        )

    def install_new_aur_deps(self) -> None:
        new_aur_deps_to_install = {
            pkg_name: self.package_builds_by_name[pkg_name].built_packages_paths[pkg_name]
            for pkg_name in self.aur_deps_names
        }
        try:
            install_built_deps(
                deps_names_and_paths=new_aur_deps_to_install,
                resolved_conflicts=self.resolved_conflicts,
            )
        except DependencyError as exc:
            if not ask_to_continue(default_yes=False):
                self._revert_transaction(PackageSource.AUR)
                raise SysExit(125) from exc
        else:
            self._save_transaction(
                PackageSource.AUR, installed=list(new_aur_deps_to_install.keys()),
            )

    def install_aur_packages(self) -> None:
        aur_packages_to_install = {
            pkg_name: self.package_builds_by_name[pkg_name].built_packages_paths[pkg_name]
            for pkg_name in self.aur_packages_names
        }
        if aur_packages_to_install:
            if not retry_interactive_command(
                    sudo([
                        *get_pacman_command(),
                        "--upgrade",
                        *reconstruct_args(
                            self.args,
                            ignore_args=["upgrade", "sync", "sysupgrade", "refresh", "ignore"],
                        ),
                        *[str(path) for path in aur_packages_to_install.values()],
                    ]),
                    pikspect=True,
                    conflicts=self.resolved_conflicts,
            ) and not ask_to_continue(default_yes=False):  # pragma: no cover
                self._revert_transaction(PackageSource.AUR)
                raise SysExit(125)
            PackageDB.discard_local_cache()
            self._save_transaction(
                PackageSource.AUR, installed=list(aur_packages_to_install.keys()),
            )

    def install_packages(self) -> None:

        if not self.args.aur:
            self.install_repo_packages()

        self.build_packages()
        if (
                not self.args.downloadonly
        ) and (
            not self.args.pkgbuild or self.args.install
        ):
            self.install_new_aur_deps()
            self.install_aur_packages()

        # save git hash of last successfully installed package
        if self.package_builds_by_name:
            package_builds_by_base = {
                pkgbuild.package_base: pkgbuild
                for pkgbuild in self.package_builds_by_name.values()
            }
            for package_build in package_builds_by_base.values():
                if len(package_build.built_packages_paths) == len(package_build.package_names):
                    if not self.args.downloadonly:
                        package_build.update_last_installed_file()
                    if not package_build.keep_build_dir:
                        remove_dir(package_build.build_dir)

        if self.failed_to_build_package_names:
            print_stderr("\n".join([
                color_line(translate("Failed to build following packages:"), ColorsHighlight.red),
                *self.failed_to_build_package_names,
            ]))
            raise SysExit(1)



import fcntl
import time
from pathlib import Path
from typing import TYPE_CHECKING

from .core import DEFAULT_INPUT_ENCODING
from .logging import create_logger

if TYPE_CHECKING:
    from typing import Any, Final, TextIO


logger_no_lock = create_logger("FileLock", lock=False)


LOCK_CHECK_INTERVAL: "Final" = 0.01  # seconds


class FileLock():

    locked = False
    lock_file: "TextIO | None"

    def __init__(self, lock_file_path: str | Path) -> None:
        self.lock_file_path = Path(lock_file_path)

    def __enter__(self) -> None:
        logger_no_lock.debug(
            "Acquiring {lock_file}...",
            lock_file=self.lock_file_path,
        )
        self.lock_file = self.lock_file_path.open("a", encoding=DEFAULT_INPUT_ENCODING)
        already_logged = False
        while True:
            try:
                fcntl.flock(self.lock_file, fcntl.LOCK_EX | fcntl.LOCK_NB)
                self.locked = True
                break
            except BlockingIOError as err:
                if not already_logged:
                    logger_no_lock.debug(
                        "Can't lock {lock_file}: {reason}",
                        lock_file=self.lock_file_path,
                        reason=err.strerror,
                    )
                    already_logged = True
                time.sleep(LOCK_CHECK_INTERVAL)
        logger_no_lock.debug(
            "Acquired {lock_file}",
            lock_file=self.lock_file_path,
        )

    def __exit__(self, *_exc_details: "Any") -> None:
        if self.lock_file:
            logger_no_lock.debug(
                "Releasing {lock_file}",
                lock_file=self.lock_file_path,
            )
            if self.locked:
                fcntl.flock(self.lock_file, fcntl.LOCK_UN)
            if not self.lock_file.closed:
                self.lock_file.close()
            self.lock_file = None
            self.locked = False
            if self.lock_file_path.exists():
                self.lock_file_path.unlink()
                logger_no_lock.debug(
                    "Released {lock_file}",
                    lock_file=self.lock_file_path,
                )

    def __del__(self) -> None:
        self.__exit__()
